import random
from typing import Type

import numpy as np
import xarray as xr

import emiproc
from emiproc.profiles.temporal_profiles import (
    AnyTimeProfile,
    CompositeTemporalProfiles,
    DailyProfile,
    DayOfLeapYearProfile,
    DayOfYearProfile,
    MounthsProfile,
    TemporalProfile,
    WeeklyProfile,
    from_csv,
    read_temporal_profiles,
)

copernicus_profiles_dir = emiproc.FILES_DIR / "profiles" / "copernicus"

TEST_COPENICUS_PROFILES = ["hour_in_day", "day_in_week", "month_in_year"]


def read_test_copernicus() -> dict[str, list[AnyTimeProfile]]:
    """Read the test copernicus profiles."""

    return {
        p: from_csv(copernicus_profiles_dir / f"timeprofiles-{p}.csv")
        for p in TEST_COPENICUS_PROFILES
    }


def get_random_profiles(
    num: int,
    profile_types: list[Type[TemporalProfile]] = [
        DailyProfile,
        WeeklyProfile,
        MounthsProfile,
    ],
) -> list[list[AnyTimeProfile]]:
    """Get random profiles for testing."""
    return [
        [
            type(ratios=((ratios := np.random.rand(type.size)) / ratios.sum()))
            for type in profile_types
            if random.choice([True, False])
        ]
        for _ in range(num)
    ]


three_profiles = [
    [
        WeeklyProfile(ratios=[0.1, 0.2, 0.3, 0.1, 0.15, 0.05, 0.1]),
        MounthsProfile(
            ratios=[0.25, 0.02, 0.03, 0.01, 0.015, 0.005, 0.11, 0.01, 0, 0.3, 0.1, 0.15]
        ),
    ],
    [
        WeeklyProfile(ratios=[0.1, 0.2, 0.3, 0.1, 0.15, 0.05, 0.1]),
        MounthsProfile(
            ratios=[
                0.01,
                0.02,
                0.03,
                0.01,
                0.015,
                0.005,
                0.11,
                0.01,
                0.24,
                0.3,
                0.1,
                0.15,
            ]
        ),
    ],
    [
        WeeklyProfile(ratios=[0.3, 0.2, 0.3, 0.1, 0.05, 0.05, 0.0]),
        MounthsProfile(
            ratios=[
                0.01,
                0.02,
                0.03,
                0.01,
                0.015,
                0.005,
                0.11,
                0.01,
                0.24,
                0.3,
                0.1,
                0.15,
            ]
        ),
    ],
]
three_composite_profiles = CompositeTemporalProfiles(three_profiles)


oem_const_profile = [
    DailyProfile(),
    MounthsProfile(),
    WeeklyProfile(),
]
weekly_test_profile = WeeklyProfile(ratios=[0.11, 0.09, 0.10, 0.09, 0.14, 0.24, 0.23])
mounths_test_profile = MounthsProfile(
    ratios=[0.19, 0.17, 0.13, 0.06, 0.05, 0.00, 0.00, 0.00, 0.01, 0.04, 0.15, 0.20]
)
# 24 hours
daily_test_profile = DailyProfile(
    ratios=[
        0.02667,
        0.01667,
        0.00667,
        0.00667,
        0.01667,
        0.02467,
        0.02700,
        0.05000,
        0.06250,
        0.06458,
        0.06250,
        0.05417,
        0.04583,
        0.04375,
        0.04375,
        0.04167,
        0.04167,
        0.04167,
        0.05000,
        0.05833,
        0.06250,
        0.05624,
        0.05416,
        0.04166,
    ]
)
day_of_year_ratios = np.arange(365, dtype=float)
day_of_year_test_profile = DayOfYearProfile(
    day_of_year_ratios / day_of_year_ratios.sum()
)
oem_test_profile = [
    weekly_test_profile,
    mounths_test_profile,
    daily_test_profile,
]

# For the inventories in test_inventories.py

indexes_inv_catsub = xr.DataArray(
    data=np.array(
        [
            [0, 1, 2],
            [3, 2, -1],
        ]
    ),
    dims=["category", "substance"],
    coords={
        "category": ["adf", "liku"],  # omit one category on purpose
        "substance": ["CH4", "CO2", "NH3"],
    },
)
indexes_inv_catsub_missing = xr.DataArray(
    data=np.array(
        [
            [0, 1],
        ]
    ),
    dims=["category", "substance"],
    # omit many on purpose
    coords={
        "category": ["adf"],
        "substance": ["CH4", "CO2"],
    },
)
indexes_inv_catsubcell = xr.DataArray(
    data=np.array(
        [
            [[1, 1, 1, 1, 1], [-1, -1, -1, 0, 1], [-1, -1, -1, 0, 2]],
            [[0, 0, 1, 2, -1], [2, 2, 2, 2, 2], [-1, -1, 0, 2, 1]],
        ]
    ),
    dims=["category", "substance", "cell"],
    coords={
        "category": ["adf", "liku"],  # omit one category on purpose
        "substance": ["CH4", "CO2", "NH3"],
        "cell": np.arange(5),
    },
)

# For the african test set
african_countries_test_set = ["SEN", "MLI", "MRT", "GIN", "GNB", "LBR", "SLE", "GMB"]
indexes_african_simple = xr.DataArray(
    data=np.arange(len(african_countries_test_set)),
    dims=["country"],
    coords={"country": african_countries_test_set},
)
indexes_african_2d = xr.DataArray(
    data=np.arange(len(african_countries_test_set) * 3).reshape(
        (len(african_countries_test_set), 3)
    ),
    dims=["country", "category"],
    coords={
        "country": african_countries_test_set,
        "category": ["liku", "blek", "test"],
    },
)



import os
from pathlib import Path

import geopandas as gpd
import numpy as np
import pytest
import xarray as xr

from emiproc.exports.wrf import WRF_Grid, export_wrf_hourly_emissions
from emiproc.inventories import Inventory
from emiproc.tests_utils import TEST_OUTPUTS_DIR, TESTFILES_DIR, temporal_profiles

test_grid_filepath = Path(TESTFILES_DIR / "wrf" / "wrfinput_for_testing.nc")


def test_grid_file_is_there():
    assert test_grid_filepath.is_file()


def test_load_grid():
    grid = WRF_Grid(test_grid_filepath)

    assert grid is not None
    assert isinstance(grid, WRF_Grid)

    assert grid.shape == (3, 6)


def get_inventory():
    grid = WRF_Grid(test_grid_filepath)
    gdf = grid.gdf.copy()

    inv = Inventory.from_gdf(
        gpd.GeoDataFrame(
            {
                ("CO2", "test"): np.ones(len(gdf)),
                ("CO2", "test2"): np.arange(len(gdf)),
            },
            geometry=gdf.geometry,
        )
    )
    inv.grid = grid

    return inv


def get_test_file_path() -> tuple[Path, Path]:
    output_first_file = TEST_OUTPUTS_DIR / "wrf" / "wrfchemi_d01_2018-01-01_00:00:00"
    output_last_file = TEST_OUTPUTS_DIR / "wrf" / "wrfchemi_d01_2018-01-01_23:00:00"

    # Windows has no `:` in filenames
    if os.name == "nt":
        replace = lambda x: x.parent / x.name.replace(":", "-")
        output_first_file = replace(output_first_file)
        output_last_file = replace(output_last_file)

    return output_first_file, output_last_file


@pytest.mark.incremental
class TestOutputs:

    def test_inventory_export(self):
        inv = get_inventory()
        inv.set_profiles(
            [[temporal_profiles.daily_test_profile]],
            indexes=xr.DataArray([0], coords={"substance": ["CO2"]}),
        )

        output_dir = TEST_OUTPUTS_DIR / "wrf"
        output_dir.mkdir(exist_ok=True, parents=True)

        export_wrf_hourly_emissions(
            inv, inv.grid, ("2018-01-01", "2018-01-02"), output_dir
        )

    def test_output_file(self):
        output_first_file, output_last_file = get_test_file_path()

        assert output_first_file.is_file()
        assert output_last_file.is_file()

    def test_file_content(self):
        file_path, _ = get_test_file_path()

        ds = xr.open_dataset(file_path)

        assert ds is not None
        assert isinstance(ds, xr.Dataset)

        assert "E_test_CO2" in ds
        assert "E_test2_CO2" in ds

        assert "Times" in ds
        assert ds["Times"].shape == (1,)
        assert ds["Times"].values[0] == b"2018-01-01_00:00:00"

        # Requires many attributes, just test one or two randomly
        assert len(ds.attrs) > 20
        assert "GFDDA_INTERVAL_M" in ds.attrs
        assert "DX" in ds.attrs
        assert "emiproc" in ds.attrs



"""Test file for tno inventory"""

from pathlib import Path

import pytest

import emiproc
from emiproc.inventories.tno import TNO_Inventory
from emiproc.inventories.utils import group_categories
from emiproc.profiles.vertical_profiles import check_valid_vertical_profile
from emiproc.utilities import get_country_mask

# TODO change the pth of that if you want to test it
tno_path = emiproc.FILES_DIR / "TNO_6x6_GHGco_v4_0/TNO_GHGco_v4_0_year2018.nc"


tno_template = emiproc.FILES_DIR / "test/tno/tno_test_minimal.nc"


def test_loading_template():
    """Test loading the template"""
    inv_tno = TNO_Inventory(tno_template)


# make this test only if the tno inventory is available
# otherwise skip it
@pytest.mark.slow
def test_loading_and_grouping():
    # Test vertical profiles on the TNO inventory
    if not tno_path.exists():
        raise FileNotFoundError(
            f"File {tno_path} not found, please add it to {tno_path}"
        )

    inv_tno = TNO_Inventory(tno_path)
    # Check the vertical profiles
    check_valid_vertical_profile(inv_tno.v_profiles)

    groupped_tno = group_categories(inv_tno, {"all": inv_tno.categories})
    check_valid_vertical_profile(groupped_tno.v_profiles)
    # test that we have the same number of point sources in both inventories
    # the number of point source is the number of rows of each of the gdfs

    assert sum([len(gdf) for gdf in inv_tno.gdfs.values()]) == len(
        groupped_tno.gdfs["all"]
    )


def test_country_ratios():
    inv_tno = TNO_Inventory(tno_template)
    # test getting the country mask of the test grid of tno
    ratios = get_country_mask(inv_tno.grid, return_fractions=True)


def test_with_t_profiles():
    inv_tno = TNO_Inventory(
        tno_template, profiles_dir=emiproc.FILES_DIR / "test/tno/profiles_cat"
    )
    assert "category" in inv_tno.t_profiles_indexes.dims
    assert "substance" not in inv_tno.t_profiles_indexes.dims


def test_with_t_profiles_catsub():
    inv_tno = TNO_Inventory(
        tno_template, profiles_dir=emiproc.FILES_DIR / "test/tno/profiles_catsub"
    )
    assert "category" in inv_tno.t_profiles_indexes.dims
    assert "substance" in inv_tno.t_profiles_indexes.dims


def test_with_t_profiles_misssing():
    inv_tno = TNO_Inventory(
        tno_template, profiles_dir=emiproc.FILES_DIR / "test/tno/profiles_missing_cat"
    )
    assert "category" in inv_tno.t_profiles_indexes.dims
    assert "substance" not in inv_tno.t_profiles_indexes.dims


def test_with_t_profiles_catsub_misssing():
    inv_tno = TNO_Inventory(
        tno_template,
        profiles_dir=emiproc.FILES_DIR / "test/tno/profiles_catsub_missing",
    )
    assert "category" in inv_tno.t_profiles_indexes.dims
    assert "substance" in inv_tno.t_profiles_indexes.dims
    missing_cat = "F2"
    other_cats = ["A", "B", "F1"]
    assert all(inv_tno.t_profiles_indexes.sel(category=missing_cat).values == -1)
    assert all(
        [cat in inv_tno.t_profiles_indexes["category"].values for cat in other_cats]
    )


if __name__ == "__main__":
    pytest.main([__file__])



"""Mapping from zh categories to GNRF sectors."""

ZH_2_GNFR = {
    # PublicPower
    "GNFR_A": [
        "c2201_BHKW_Emissionen_Kanton",
        "c2301_KHKWKehricht_Emissionen_Kanton",
        "c2302_KHKWErdgas_Emissionen_Kanton",
        "c2303_KHKWHeizoel_Emissionen_Kanton",
    ],
    # Industry
    "GNFR_B": [
        "c3201_Notstromanlagen_Emissionen_Kanton",
        "c3301_Prozessenergie_Emissionen_Kanton",
        "c3401_Metallreinigung_Emissionen_Kanton",
        "c3402_Holzbearbeitung_Emissionen_Kanton",
        "c3403_Malereien_Emissionen_Kanton",
        "c3404_Textilreinigung_Emissionen_Kanton",
        "c3405_Karosserien_Emissionen_Kanton",
        "c3406_Raeuchereien_Emissionen_Kanton",
        "c3407_Roestereien_Emissionen_Kanton",
        "c3408_Druckereien_Emissionen_Kanton",
        "c3409_Laboratorien_Emissionen_Kanton",
        "c3410_Bierbrauereien_Emissionen_Kanton",
        "c3411_Brotproduktion_Emissionen_Kanton",
        "c3412_MedizinischePraxen_Emissionen_Kanton",
        "c3413_Gesundheitswesen_Emissionen_Kanton",
    ],
    # Other stationary combustion (services, residential, agriculture)
    "GNFR_C": [
        "c2101_Oelheizungen_Emissionen_Kanton",
        "c2102_Gasheizungen_Emissionen_Kanton",
        "c2103_HolzheizungenLokalisiert_Emissionen_Kanton",
        "c2104_HolzheizungenDispers_Emissionen_Kanton",
        "c2105_Warmwassererzeuger_Emissionen_Kanton",
    ],
    # Fugitives
    "GNFR_D": [
        "c3416_Tankstellen_Emissionen_Kanton",
    ],
    # Solvents and product use
    "GNFR_E": [
        "c3417_LoesemittelIG_Emissionen_Kanton",
        "c5101_LoesemittelHH_Emissionen_Kanton",
    ],
    # Road transport
    "GNFR_F": [
        "c1301_Personenwagen_Emissionen_Kanton",
        "c1302_Lastwagen_Emissionen_Kanton",
        "c1303_Motorraeder_Emissionen_Kanton",
        "c1304_Linienbusse_Emissionen_Kanton",
        "c1305_Trolleybusse_Emissionen_Kanton",
        "c1306_StartStopTankatmung_Emissionen_Kanton",
        "c1307_Lieferwagen_Emissionen_Kanton",
        "c1308_Reisebusse_Emissionen_Kanton",
    ],
    # Shipping
    "GNFR_G": [
        "c1101_Linienschiffe_Emissionen_Kanton",
        "c1102_PrivaterBootsverkehr_Emissionen_Kanton",
    ],
    # Offroad mobility
    "GNFR_I": [
        "c1201_BahnPersonenverkehr_Emissionen_Kanton",
        "c1202_BahnGueterverkehr_Emissionen_Kanton",
        "c1203_Tramverkehr_Emissionen_Kanton",
        "c1204_Kleinbahnen_Emissionen_Kanton",
        # c31xx are construction stuff
        "c3101_MaschinenHochbau_Emissionen_Kanton",
        "c3102_Bitumen_Emissionen_Kanton",
        "c3103_FarbenBaustelle_Emissionen_Kanton",
        "c3104_MaschinenTiefbau_Emissionen_Kanton",
        "c3105_Strassenbelag_Emissionen_Kanton",
        "c3419_IndustrielleFZ_Emissionen_Kanton",
        "c4101_ForstwirtschaftlicheFZ_Emissionen_Kanton",
        "c4201_LandwirtschaftlicheFZ_Emissionen_Kanton",
    ],
    # Waste
    "GNFR_J": [
        "c2401_Klaerschlammverwertung_Emissionen_Kanton",
        "c3418_Vergaerwerk_Emissionen_Kanton",
        "c3414_Krematorium_Emissionen_Kanton",
        "c5201_Gruenabfallverbrennung_Emissionen_Kanton",
        "c5301_HolzoefenKleingarten_Emissionen_Kanton",
        "c5401_AbfallverbrennungHaus_Emissionen_Kanton",
    ],
    # AgriLivestock
    "GNFR_K": [
        "c4401_Nutztierhaltung_Emissionen_Kanton",
    ],
    # AgriOther
    "GNFR_L": [
        "c4301_Nutzflaechen_Emissionen_Kanton",
    ],
    # Others
    "GNFR_R": [
        "c5501_HausZooZirkustiere_Emissionen_Kanton",
        "c5601_Feuerwerke_Emissionen_Kanton",
        "c5701_Tabakwaren_Emissionen_Kanton",
        "c5801_BrandFeuerschaeden_Emissionen_Kanton",
        "c6101_Waelder_Emissionen_Kanton",
        "c6201_Grasflaechen_Emissionen_Kanton",
        "c6301_Gewaesser_Emissionen_Kanton",
        "c6401_Blitze_Emissionen_Kanton",
    ],
}



import urllib.request
from datetime import date, datetime
from os import PathLike
from pathlib import Path

import geopandas as gpd
import numpy as np
import xarray as xr

from emiproc.grids import Grid, RegularGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    DayOfYearProfile,
    Hour3OfDayPerMonth,
    MounthsProfile,
    get_leap_year_or_normal,
)
from emiproc.profiles.utils import ratios_dataarray_to_profiles


def download_gfed5(
    data_dir: PathLike,
    year: int,
    link_template: str = "https://surfdrive.surf.nl/files/index.php/s/VPMEYinPeHtWVxn/download?path=%2Fdaily&files=GFED5_Beta_daily_{year}{month}.nc",
):
    """Download the GFED5 files for a given year.

    The files are downloaded in the data_dir folder.

    :param data_dir: The directory where to download the files
    :param year: The year to download
    :param link_template: The template for the download link. The template should contain the year and month placeholders.
    """

    data_dir = Path(data_dir)

    for month in range(1, 13):
        link = link_template.format(year=year, month=f"{month:02d}")
        filename = link.split("=")[-1]
        filepath = data_dir / filename
        try:
            urllib.request.urlretrieve(link, filepath)
        except urllib.error.HTTPError as e:
            raise ValueError(
                f"Link {link} does not exist. Maybe the inventory is not available for {year=}."
            ) from e

    print(f"Downloaded gfed5 files for {year=}.")


class GFED_Grid(RegularGrid):
    """Grid for the GFED inventory.


    The hdf5 file contains all the information about the grid coordinates.
    It is a regular grid.


    Not also that the grid cell area is provided in the file,
    but we do not use it here.
    Also the values are not exactly the same (around 10% errors).
    Geopandas calculated areas are a bit larger it seems.

    .. note::

        The profiles implemented are only the day of year profile.
        If you want hour of day, you will have to implement it.

    """

    def __init__(self, gfed_filepath: PathLike):

        gfed_filepath = Path(gfed_filepath)
        ds = xr.open_dataset(gfed_filepath)

        # Get the lon lat coordinates
        # This assumes (but checks) that the grid is regular
        unique_lons = np.unique(ds["lon"].values, axis=0)
        unique_lats = np.unique(ds["lat"].values.T, axis=0)
        assert len(unique_lons) == 1
        assert len(unique_lats) == 1

        self.lon_range = unique_lons[0]
        self.lat_range = unique_lats[0]

        # Get the grid cell size (here also ensure that the grid is regular)
        unique_dx = np.unique(np.diff(self.lon_range))
        unique_dy = np.unique(np.diff(self.lat_range))

        assert len(unique_dx) == 1
        assert len(unique_dy) == 1

        self.dx = abs(unique_dx[0])
        self.dy = abs(unique_dy[0])

        self.nx = len(self.lon_range)
        self.ny = len(self.lat_range)

        # Bypass the RegularGrid __init__ method because we already have the grid coordinates
        Grid.__init__(self, gfed_filepath.stem)


class GFED4_Inventory(Inventory):
    """Global Fire Emissions Database.

    Global inventory based on satellite data, burned areas, fuel consumption.

    https://www.globalfiredata.org/

    You can download the input data for various year at
    https://www.geo.vu.nl/~gwerf/GFED/GFED4/

    This uses the area inside the GFED file to calculate the total emissions, but
    we found out that the area is not exactly the same as the one calculated by geopandas.
    See :py:class:`GFED_Grid` for the grid information.

    The data set contains two variables:
        * C: Carbon emissions
        * DM: Dry matter emissions

    .. note:: This inventory applies only for GFED4 .
        GFED5 has changed the format and is not supported by this class.
    """

    def __init__(self, gfed_filepath: PathLike, year: int):

        super().__init__()

        self.gfed_filepath = Path(gfed_filepath)
        gfed_file = self.gfed_filepath
        self.grid = GFED_Grid(gfed_filepath)

        self.year = year

        # Units of C var: g C / m^2 / month
        # Units of DM var: kg DM / m^2 / month

        das = []
        phony_dims = lambda ds: [dim for dim in ds.dims if dim.startswith("phony_dim")]
        lat_lon_dims = lambda ds: {phony_dims(ds)[0]: "lat", phony_dims(ds)[1]: "lon"}
        rename_phony_dims = lambda ds: ds.rename(lat_lon_dims(ds))
        for month in range(1, 13):
            da_dm = xr.open_dataset(gfed_file, group=f"/emissions/{month:02}")["DM"]
            ds_partion = xr.open_dataset(
                gfed_file, group=f"/emissions/{month:02}/partitioning"
            )
            # Get teh phony dims and renmae them
            ds_partion = rename_phony_dims(ds_partion)
            da_dm = rename_phony_dims(da_dm)
            da_partition = ds_partion.to_dataarray(dim="category")
            das.append((da_dm * da_partition).expand_dims(month=[month]))
        da = xr.concat(das, dim="month")

        # Rename the category to remove the `DM_` prefix
        da["category"] = [str(cat).split("_")[-1] for cat in da["category"].values]

        # Get the grid cell areas
        grid_areas = xr.open_dataset(gfed_file, group="/ancill/")["grid_cell_area"]
        grid_areas = rename_phony_dims(grid_areas)
        # Scale with the grid cell area to get kg / year / cell
        da = da * grid_areas

        da_stacked = da.stack(cell=("lon", "lat")).drop_vars(["lon", "lat"])
        cell_index = np.array(range(da_stacked.sizes["cell"]))
        da_stacked = da_stacked.assign_coords(cell=cell_index)

        da_total = da_stacked.sum(dim="month")

        mask_cell = da_total.sum(dim="category") > 0

        columns = {}

        for category in da_total["category"].values:
            columns[(category, "DM")] = da_total.sel(category=category).values

        self.gdf = gpd.GeoDataFrame(columns, geometry=self.grid.gdf.geometry)
        self.gdfs = {}

        # Now we make the profiles

        das_daily = []
        year = 2018
        for month in range(1, 13):
            ds_daily = xr.open_dataset(
                gfed_file, group=f"/emissions/{month:02}/daily_fraction"
            )

            # Merge the days into a new dimension
            da_daily = xr.concat(
                [
                    ds_daily[f"day_{i:d}"].expand_dims(
                        day=[datetime(year=year, month=month, day=i)]
                    )
                    for i in range(1, 32)
                    if f"day_{i:d}" in ds_daily
                ],
                dim="day",
            )
            da_daily = rename_phony_dims(da_daily)
            das_daily.append(da_daily)

        da_day_of_year = xr.concat(das_daily, dim="day")
        da_day_of_year = da_day_of_year / da_day_of_year.sum(dim="day")

        # Stack to have on the cell dimension
        da_day_of_year_stacked = (
            da_day_of_year.stack(cell=("lon", "lat"))
            .drop_vars(["lon", "lat"])
            .assign_coords(cell=cell_index)
        )

        dss_diurnal = []
        for month in range(1, 13):
            ds_diurnal = xr.open_dataset(
                gfed_file, group=f"/emissions/{month:02}/diurnal_cycle"
            )
            dss_diurnal.append(rename_phony_dims(ds_diurnal).expand_dims(month=[month]))
        ds_diurnal = xr.concat(dss_diurnal, dim="month")
        da_diurnal = xr.concat(
            [
                ds_diurnal[hour_name]
                for hour_name in [
                    "UTC_0-3h",
                    "UTC_3-6h",
                    "UTC_6-9h",
                    "UTC_9-12h",
                    "UTC_12-15h",
                    "UTC_15-18h",
                    "UTC_18-21h",
                    "UTC_21-24h",
                ]
            ],
            dim="hour",
        )
        da_month_diurnal = xr.concat(
            [
                da_diurnal.sel(month=month).assign_coords(
                    hour=8 * (month - 1) + np.arange(8)
                )
                for month in range(1, 13)
            ],
            dim="hour",
        ).rename(hour="hour3_per_month")
        da_diurnal3_stacked = (
            da_month_diurnal.stack(cell=("lon", "lat"))
            .drop_vars(["lon", "lat"])
            .assign_coords(cell=cell_index)
        )

        # Below whe should use the function that generates composite profiles

        profiles_arrays = {
            Hour3OfDayPerMonth: da_diurnal3_stacked.rename(
                {"hour3_per_month": "ratio"}
            ).drop_vars("month"),
            DayOfYearProfile: da_day_of_year_stacked.rename({"day": "ratio"}),
            MounthsProfile: da_stacked.rename({"month": "ratio"}),
        }

        parsed_profiles = {
            profile_type: (
                # Convert to scaling factors to ease nan handling
                da
                / da.mean(dim="ratio")
            )
            .fillna(1.0)
            .sel(cell=mask_cell)
            # Add the category dim
            .broadcast_like(da_stacked, exclude=["cell", "month"])
            # Drop the ratio coordinate values, so we can concat on the ratio dimension
            .drop_vars("ratio")
            for profile_type, da in profiles_arrays.items()
        }

        # Set the ratios all together and build the composite profiles
        profiles_ratios = xr.concat(parsed_profiles.values(), dim="ratio")
        ratios, indices = ratios_dataarray_to_profiles(profiles_ratios)

        profiles = CompositeTemporalProfiles.from_ratios(
            ratios, list(parsed_profiles.keys()), rescale=True
        )

        self.set_profiles(profiles, indices)


class GFED5(Inventory):
    """Global Fire Emissions Database.

    Global inventory based on satellite data, burned areas, fuel consumption.

    https://www.globalfiredata.org/

    You can download the input data for various year using :py:func:`download_gfed5`.
    """

    def __init__(self, file_dir: PathLike, year: int, substances: list[str]):
        super().__init__()

        files_dir = Path(file_dir)
        files = [
            files_dir / f"GFED5_Beta_daily_{year}{month:02d}.nc"
            for month in range(1, 13)
        ]

        # Check that all files exists
        for file in files:
            if not file.exists():
                raise ValueError(f"File {file} does not exist.")

        ds = xr.open_mfdataset(files, combine="by_coords")

        self.grid = RegularGrid.from_centers(
            x_centers=ds["lon"].values,
            y_centers=ds["lat"].values,
            name="GFED5",
        )

        profiles = {}

        for sub in substances:
            if sub not in ds.data_vars:
                raise ValueError(f"Substance {sub} not in the dataset.")

            profiles[sub] = ds[sub].expand_dims(substance=[sub])

        da_profiles: xr.DataArray = (
            xr.concat(list(profiles.values()), dim="substance")
            .stack(cell=("lon", "lat"))
            .drop_vars(["lon", "lat"])
            .rename({"time": "ratio"})
        )

        self.gdf = gpd.GeoDataFrame(
            {
                ("gfed", sub): da_profiles.sel(substance=sub).sum(dim=["ratio"]).values
                # Convert from kg/m2 to kg/cell
                * 1e-3 * self.grid.cell_areas
                for sub in substances
            },
            geometry=self.grid.gdf.geometry,
        )
        self.gdfs = {}

        ratios, indices = ratios_dataarray_to_profiles(da_profiles)

        profiles = CompositeTemporalProfiles.from_ratios(
            ratios, [get_leap_year_or_normal(DayOfYearProfile, year=year)], rescale=True
        )

        self.set_profiles(profiles, indices)



from __future__ import annotations
import logging
from os import PathLike
from pathlib import Path
import xarray as xr
import numpy as np
from emiproc.inventories import Inventory
from emiproc.grids import RegularGrid
from emiproc.profiles.utils import get_desired_profile_index
from emiproc.exports.netcdf import NetcdfAttributes
from emiproc.utilities import (
    get_timezone_mask,
)

from emiproc.exports.icon import (
    TemporalProfilesTypes,
    make_icon_time_profiles,
    make_icon_vertical_profiles,
)


def export_inventory_profiles(
    inv: Inventory,
    output_dir: PathLike,
    grid: RegularGrid,
    netcdf_attributes: NetcdfAttributes,
    var_name_format: str = "{substance}_{category}",
    temporal_profiles_type: TemporalProfilesTypes = TemporalProfilesTypes.THREE_CYCLES,
    year: int | None = None,
) -> Path:
    """Export the vertical and temporal profiles of the inventory to netcdfs.

    This will export the vertical and temporal profiles of the inventory to netcdf files.
    The vertical profiles will be exported to a single netcdf file.
    The temporal profiles will be exported according to
    :py:func:`~emiproc.exports.icon.make_icon_time_profiles`.

    :param inv: the inventory to export
    :param output_dir: the path to the output directory
    :param grid: the raster grid to export to.
        This is used to calculate the country shifts for the temporal profiles.
    :param netcdf_attributes: Same as in :py:func:`emiproc.exports.netcdf.export_raster_netcdf`.
    :param var_name_format: Same as in :py:func:`emiproc.exports.netcdf.export_raster_netcdf`.
    :param temporal_profiles_type: The type of temporal profiles to export.
        See :py:class:`emiproc.exports.icon.TemporalProfilesTypes` for the available options.
    :param year: The year of the temporal profiles.
    """
    logger = logging.getLogger("emiproc.export_inventory_profiles")
    output_dir = Path(output_dir)

    if "type" in inv.v_profiles_indexes.coords:
        raise NotImplementedError(
            "Can only export only profiles varying on 'category' and 'substance' . \n"
            "If you want to export inventory profiles, "
            "please use another function or remap your inventory."
        )

    # Add the vertical profiles
    if inv.v_profiles:
        profiles = {
            var_name_format.format(substance=sub, category=cat): inv.v_profiles[
                get_desired_profile_index(inv.v_profiles_indexes, cat=cat, sub=sub)
            ]
            for cat in inv.categories
            for sub in inv.substances
        }
        make_icon_vertical_profiles(
            profiles,
            output_dir,
            nc_attrs=netcdf_attributes,
        )
    else:
        logger.warning(
            f"No vertical profiles found in {inv}, no vertical profiles will be"
            " exported."
        )

    if inv.t_profiles_groups:
        tz_mask = get_timezone_mask(grid)
        profiles = {
            var_name_format.format(substance=sub, category=cat): inv.t_profiles_groups[
                get_desired_profile_index(inv.t_profiles_indexes, cat=cat, sub=sub)
            ]
            for cat in inv.categories
            for sub in inv.substances
        }

        unique_tz, tz_indexes = np.unique(tz_mask, return_inverse=True)

        make_icon_time_profiles(
            time_profiles=profiles,
            time_zones=unique_tz,
            profiles_type=temporal_profiles_type,
            year=year,
            out_dir=output_dir,
            nc_attrs=netcdf_attributes,
        )
        ds_tz_mask = xr.Dataset(
            {
                "tz_mask": (
                    ["lat", "lon"],
                    tz_indexes.reshape(grid.shape).T,
                ),
                "timezones": (
                    ["country_id"],
                    unique_tz,
                ),
            },
            coords={
                "lat": grid.lat_range,
                "lon": grid.lon_range,
            },
            attrs=netcdf_attributes,
        )
        ds_tz_mask.to_netcdf(output_dir / "tz_mask.nc")

    else:
        logger.warning(
            f"No temporal profiles found in {inv}, no temporal profiles will be"
            " exported."
        )



"""Emission generation for the GRAL model: https://gral.tugraz.at/ .

This module contains functions to prepare emissions for GRAL.

Gral only has one polluant available, but can contain more than one category.

To counter that issue, we make a source group for each substance/category.


.. warning:: For PM, it seems that there are some additional parameters:
    share_pm25,share_pm10,diamter_pm30[μm],density[kg/m3],dry_dep_vel_pm25[m/s],
    dry_dep_vel_pm10[m/s],dry_dep_vel_pm30[m/s],mode

    We don't use them for now.

.. warning:: The current version of emiproc does not support lines and portals.


4 types of emissions are supported with their respective files:

- point sources: point.dat
- line sources: line.dat
- area sources: cadastre.dat
- tunnel sources: portals.dat

Each file is optional.


The following methods are applied:

- points: simply write the coordinates and the emission rate
- lines: write the coordinates of the line and the emission rate
    The Multi-lines have to be split into single lines, which increases the
    size of the problem.
- areas: polygons have to be rasterized into squares before writing them.
- tunnels: not implemented 



"""

from __future__ import annotations

import datetime
import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

import geopandas as gpd
import numpy as np
import pandas as pd
from rasterio.enums import MergeAlg
from rasterio.features import rasterize
from rasterio.transform import from_bounds
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Point, Polygon

from emiproc.inventories import EmissionInfo, Inventory

if TYPE_CHECKING:
    # pygg module for gram gral processing
    from pygg.grids import GralGrid


class EmissionWriter:
    def __init__(
        self,
        path: os.PathLike,
        inventory: Inventory,
        grid: GralGrid,
        # Egde size of the rasterized polygons (crs units)
        polygon_raster_size: float = 1.0,
    ) -> None:

        # Path where to write the emissons
        self.path = Path(path)

        self.inventory = inventory
        self.grid = grid
        self.polygon_raster_size = polygon_raster_size

        # Maps the (cat/sub) to source groups
        self.source_groups = {
            (sub, cat): i * len(self.inventory.categories) + j
            for i, sub in enumerate(self.inventory.substances)
            for j, cat in enumerate(self.inventory.categories)
        }

        self._make_files()

    def _make_files(self):
        """Create empty emissions files with the headers.

        Called durint init of the class.
        """
        if not self.path.is_dir():
            raise FileNotFoundError(f"Path {self.path} does not exist.")

        # Make the filepaths
        self.file_points = self.path / "point.dat"
        self.file_lines = self.path / "line.dat"
        self.file_cadastre = self.path / "cadastre.dat"
        self.file_portals = self.path / "portals.dat"

        with open(self.file_points, "w") as f:
            # 2 lines ignored
            header = datetime.datetime.now().strftime("Generated: %d/%m/%Y %H:%M\n")
            f.write(header)
        self.points_dfs = []
        with open(self.file_lines, "w") as f:
            # 5 lines ignored
            header = (
                datetime.datetime.now().strftime("Generated: %d/%m/%Y %H:%M\n")
                + 3 * "Generated: \n"
                + "Name,Section,source_group,x1,y1,z1,x2,y2,z2,width,vert. ext.,-,-,"
                "emission_rate[kg/h/km],-,-,-,-\n"
            )
            f.write(header)
        with open(self.file_cadastre, "w") as f:
            # 1 line ignored
            header = "x,y,z,dx,dy,dz,emission_rate[kg/h],-,-,-,source_group\n"
            f.write(header)
        with open(self.file_portals, "w") as f:
            # 2 lines ignored
            header = (
                datetime.datetime.now().strftime("Generated: %d/%m/%Y %H:%M\n")
                + "x1,y1,x2,y2,z0,z1,emission_rate[kg/h],-,-,-,source_group\n"
            )
            f.write(header)

        # File to save the source groups values
        self.file_source_groups = self.path / "source_groups.json"
        with open(self.file_source_groups, "w") as f:
            # reverse the dict (items become keys and vice versa)
            reversed_source_groups = {v: k for k, v in self.source_groups.items()}
            json.dump(reversed_source_groups, f, indent=2)

    def write_gdfs(self):
        """Write the gdfs emissions to the files."""
        for cat, gdf in self.inventory.gdfs.items():
            info = self.inventory.emission_infos[cat]
            for sub in self.inventory.substances:
                source_group = self.source_groups[(sub, cat)]
                if sub not in gdf.columns:
                    continue

                mask_polygons = gdf.geom_type.isin(["Polygon", "MultiPolygon"])
                if any(mask_polygons):
                    gdf_polygons = gdf.loc[mask_polygons]
                    self._write_polygons(
                        gdf_polygons.geometry, gdf_polygons[sub], info, source_group
                    )

                mask_points = gdf.geom_type == "Point"
                if any(mask_points):
                    gdf_points = gdf.loc[mask_points]
                    self._add_points(
                        gdf_points.geometry, gdf_points[sub], info, source_group
                    )

                mask_lines = gdf.geom_type.isin(["LineString"])
                if any(mask_lines):
                    gdf_lines = gdf.loc[mask_lines]
                    self._write_lines(
                        gdf_lines.geometry, gdf_lines[sub], info, source_group
                    )

                mask_multilines = gdf.geom_type.isin(["MultiLineString"])
                if any(mask_multilines):
                    gdf_multilines = gdf.loc[mask_multilines]
                    # Split all the multilines into lines
                    for shape, shape_emission in zip(
                        gdf_multilines.geometry, gdf_multilines[sub]
                    ):
                        lenghts = np.array([line.length for line in shape.geoms])
                        proprtions = lenghts / shape.length
                        for line, prop in zip(shape.geoms, proprtions):
                            self._write_line(
                                line, shape_emission * prop, info, source_group
                            )
                mask_missing = ~(
                    mask_multilines | mask_lines | mask_points | mask_polygons
                )
                if any(mask_missing):
                    raise NotImplementedError(
                        f"Shapes of type: '{gdf.loc[mask_missing].geom_type.unique()}'"
                        " are not implemented."
                    )

        # Write all the points as a singl batch
        pd.concat(self.points_dfs).to_csv(
            self.file_points,
            mode="a",
            index=False,
        )

    def _add_points(
        self,
        shapes: gpd.GeoSeries,
        emissions: pd.Series,
        info: EmissionInfo,
        source_group: int,
    ):
        z = info.height
        if info.height_over_buildings:
            z += self.grid.building_heights[self.grid.get_index(shapes.x, shapes.y)]
        self.points_dfs.append(
            pd.DataFrame(
                {
                    "x": shapes.x,
                    "y": shapes.y,
                    "z": z,
                    "emission[kg/h]": emissions,
                    # Fill the unused columsn with zeros as GRAL cannot handle empty columns
                    "unused_0": 0,
                    "unused_1": 0,
                    "unused_2": 0,
                    "exit_velocity[m/s]": info.speed,
                    "diameter[m]": info.width,
                    "temperature[K]": info.temperature,
                    "source_group": source_group,
                }
            )
        )

    def _write_lines(
        self,
        shapes: gpd.GeoSeries,
        emissions: pd.Series,
        info: EmissionInfo,
        source_group: int,
    ):
        for shape, emission in zip(shapes, emissions):
            self._write_line(shape, emission, info, source_group)

    def _write_line(
        self,
        shape: LineString,
        emission: float,
        info: EmissionInfo,
        source_group: int,
    ):
        """Write a line source to the file."""
        # split the line string in single lines
        previous_coord = shape.coords[0]

        lines = [LineString([previous_coord, coord]) for coord in shape.coords[1:]]

        # split the emission between the lines based on the length
        line_lengths = np.array([line.length for line in lines])
        line_emission_ratios = line_lengths / line_lengths.sum()
        line_emissions = line_emission_ratios * emission

        for i, (line, line_emission) in enumerate(zip(lines, line_emissions)):
            self._write_staight_line(
                Point(line.coords[0]),
                Point(line.coords[1]),
                line_emission,
                info,
                source_group,
                section=i,
            )

    def _write_staight_line(
        self,
        start: Point,
        end: Point,
        emission: float,
        info: EmissionInfo,
        source_group: int,
        section: int,  # Section number of the line
    ):
        """Write a straight line source to the file."""
        z_start = info.height
        z_end = info.height
        if info.height_over_buildings:
            z_start += self.grid.building_heights[self.grid.get_index(start.x, start.y)]
            z_end += self.grid.building_heights[self.grid.get_index(end.x, end.y)]

        with open(self.file_lines, "a") as f:
            # Write the line
            f.write(
                f"unnamed,{section},{source_group},{start.x},{start.y},{z_start},"
                f"{end.x},{end.y},{z_end},{info.width},-{info.vertical_extension},0,0,"
                f"{emission},0,0,0,0\n"
            )

    def _write_polygons(
        self,
        shapes: Iterable[Polygon],
        emissions: Iterable[float],
        info: EmissionInfo,
        source_group: int,
    ):
        """Write a polygon source to the file."""

        # Rasterize the polygon on a grid
        shapes_serie = gpd.GeoSeries(shapes)
        # get polygon bounds
        minx, miny, maxx, maxy = shapes_serie.total_bounds
        # Create a grid for the rasterization
        x = np.arange(minx, maxx, self.polygon_raster_size)
        y = np.arange(miny, maxy, self.polygon_raster_size)

        # Get the emission per cell
        average_cells_proportion = (self.polygon_raster_size**2) / shapes_serie.area
        cell_emissions = np.array(emissions) * average_cells_proportion

        # WARNING: this might be not exactly mass convserving
        rasterized_emissions = rasterize(
            shapes=zip(shapes, cell_emissions),
            out_shape=(len(x), len(y)),
            transform=from_bounds(minx, miny, maxx, maxy, len(x), len(y)),
            all_touched=False,
            merge_alg=MergeAlg.add,
        )[
            ::-1, :
        ]  # flip the y axis

        # Get the coordinates of the rasterized polygon
        indices = np.array(np.where(rasterized_emissions)).T

        # Write the polygon
        with open(self.file_cadastre, "a") as f:
            for i_x, i_y in indices:
                f.write(
                    f"{x[i_x]},{y[i_y]},{info.height},"
                    f"{self.polygon_raster_size},{self.polygon_raster_size},{info.vertical_extension},"
                    f"{rasterized_emissions[i_x, i_y]},0,0,0,{source_group},\n"
                )


def export_to_gral(
    inventory: Inventory,
    grid: GralGrid,
    path: os.PathLike,
    polygon_raster_size: float = 1.0,
) -> None:
    """Export an inventory to GRAL.

    .. note:: This requires the external python package `pygg` to be installed.

    :param inventory: Inventory to export.
    :param path: Path where to write the emissions.
    :param grid: Grid to use.
    """

    writer = EmissionWriter(Path(path), inventory, grid, polygon_raster_size)

    writer.write_gdfs()


if __name__ == "__main__":
    # Small test inventory for saving to gral format
    import logging

    import geopandas as gpd
    import numpy as np
    from shapely.geometry import (
        LineString,
        MultiLineString,
        MultiPolygon,
        Point,
        Polygon,
    )

    from emiproc.exports.gral import export_to_gral
    from emiproc.inventories import EmissionInfo, Inventory
    from emiproc.tests_utils import TEST_OUTPUTS_DIR
    from emiproc.tests_utils.test_inventories import inv_with_pnt_sources

    gral_test_inv = inv_with_pnt_sources.copy()
    # Addintg some shapes that should be supported
    gral_test_inv.gdfs["adf"] = gpd.GeoDataFrame(
        {
            "CO2": [1, 2, 1, 3, 1],
            "CH4": [3, 1, 1, 1, 1],
        },
        geometry=[
            LineString([(0.75, 0.75), (1.25, 1.25), (2.25, 1.25)]),
            Polygon([(0.25, 0.25), (0.75, 0.75), (0.25, 0.75)]),
            MultiPolygon(
                [
                    # Square from 2 to 4
                    Polygon([(2, 2), (4, 2), (4, 4), (2, 4)]),
                    # Sqare from 5 to 6
                    Polygon([(5, 5), (6, 5), (6, 6), (5, 6)]),
                    # A triangle from 8 to 12
                    Polygon([(8, 8), (12, 8), (12, 12)]),
                ]
            ),
            MultiLineString(
                [
                    LineString([(0.25, 0.25), (0.75, 0.75), (3.25, 1.25)]),
                    LineString([(0.75, 0.75), (0.5, 0.25)]),
                ]
            ),
            Point(0.5, 0.5),
        ],
    )

    from pygg.grids import GralGrid

    inv = gral_test_inv

    grid = GralGrid(
        dz0=1.5,
        stretch=0.5,
        nx=20,
        ny=20,
        nslice=20,
        sourcegroups="",
        xmin=0,
        xmax=10,
        ymin=0,
        ymax=4,
        latitude=0,
        crs=inv.crs,
    )
    grid.building_heights = np.zeros(size=(grid.ny, grid.nx))

    inv.emission_infos = {
        "adf": EmissionInfo(),
        "blek": EmissionInfo(),
        "liku": EmissionInfo(),
        "other": EmissionInfo(),
        "test": EmissionInfo(),
    }

    export_to_gral(inv, grid, TEST_OUTPUTS_DIR, 1)



"""Emission processing package."""
from __future__ import annotations

import logging
from pathlib import Path

# directory where the data files are stored
FILES_DIR = Path(__file__).parent.parent / "files"
TESTS_DIR = FILES_DIR / "test"

logger = logging.getLogger("emiproc")

# Create a dedicated logging level for processes
PROCESS = 25  # between INFO (20) and WARNING (30)
logging.addLevelName(PROCESS, "PROCESS")

logger.setLevel(PROCESS)


def deprecated(msg: str | None = None):
    """Decorator to mark functions as deprecated."""

    def deprecated_decorator(func, msg=msg):
        def wrapper(*args, msg=msg, **kwargs):
            msg_default = "Call to deprecated function {}.".format(func.__name__)
            if msg is None:
                msg = msg_default
            else:
                msg = msg_default + " " + msg
            logger.warning(msg, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return deprecated_decorator



"""Test the speciation module."""

from os import PathLike
from pathlib import Path

import xarray as xr

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from shapely.geometry import Point

from emiproc import TESTS_DIR
from emiproc.inventories import Inventory
from emiproc.profiles.vertical_profiles import VerticalProfile, VerticalProfiles
from emiproc.speciation import (
    read_speciation_table,
    speciate,
    speciate_inventory,
    speciate_nox,
    merge_substances,
)
from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources
from emiproc.tests_utils.african_case import (
    african_inv_emissions_only_land,
    african_inv,
)


@pytest.mark.parametrize(
    "table_path",
    [
        TESTS_DIR / "speciation" / "test_table_NOx.csv",
        TESTS_DIR / "speciation" / "table_test_inv_CO2.csv",
    ],
)
def test_read_speciation_table(table_path: PathLike):
    read_speciation_table(table_path)


def test_load_wrong_table():
    with pytest.raises(ValueError):
        read_speciation_table(TESTS_DIR / "speciation" / "wrong_ratio_table.csv")


def test_load_wrong_no_checks():
    read_speciation_table(
        TESTS_DIR / "speciation" / "wrong_ratio_table.csv", check_sum=False
    )


def test_speciation_wrong_ratios():
    da_speciation = read_speciation_table(
        TESTS_DIR / "speciation" / "wrong_ratio_table.csv", check_sum=False
    )
    inv_speciated = speciate(
        inv_with_pnt_sources, substance="CO2", speciation_ratios=da_speciation
    )

    # Check that the total emissions are the same
    # Ratios for adf in the file are {CO2_ANT: 0.5, CO2_BIO: 0.4}
    assert (
        inv_speciated.total_emissions.loc["CO2_ANT", "adf"]
        == 0.5 * inv_with_pnt_sources.total_emissions.loc["CO2", "adf"]
    )
    assert (
        inv_speciated.total_emissions.loc["CO2_BIO", "adf"]
        == 0.4 * inv_with_pnt_sources.total_emissions.loc["CO2", "adf"]
    )


def test_speciation_african_case():
    da_africa = read_speciation_table(
        TESTS_DIR / "speciation" / "table_africa_testcase.csv"
    )

    speciated = speciate(
        african_inv_emissions_only_land, substance="CO2", speciation_ratios=da_africa
    )

    # Test that the total emissions are the same
    pd.testing.assert_series_equal(
        speciated.total_emissions.loc[["CO2_ANT", "CO2_BIO"], :]
        .sum(axis="index")
        .sort_index(),
        african_inv_emissions_only_land.total_emissions.loc["CO2", :].sort_index(),
        check_names=False,
    )


def test_fail_on_emissions_on_no_country():
    """By default, the speciation should fail if there are emissions in cells with no country (eg on ocean)."""
    da_africa = read_speciation_table(
        TESTS_DIR / "speciation" / "table_africa_testcase.csv"
    )

    with pytest.raises(ValueError):
        speciate(
            african_inv,
            substance="CO2",
            speciation_ratios=da_africa,
        )


def test_use_default_when_country_missing():
    """Use a custom default value in the file when there are emissions in cells with no country (eg on ocean).

    The country is -99 in the file
    """
    da_africa = read_speciation_table(
        TESTS_DIR / "speciation" / "table_africa_testcase_with_missing.csv"
    )

    speciated = speciate(african_inv, substance="CO2", speciation_ratios=da_africa)

    # Test that the total emissions are the same
    pd.testing.assert_series_equal(
        speciated.total_emissions.loc[["CO2_ANT", "CO2_BIO"], :]
        .sum(axis="index")
        .sort_index(),
        african_inv.total_emissions.loc["CO2", :].sort_index(),
        check_names=False,
    )


def test_speciate_simple_inventory():
    da_speciation = read_speciation_table(
        TESTS_DIR / "speciation" / "table_test_inv_CO2.csv"
    )

    inv_speciated = speciate(
        inv_with_pnt_sources, substance="CO2", speciation_ratios=da_speciation
    )

    # Check that the total emissions are the same
    pd.testing.assert_series_equal(
        inv_speciated.total_emissions.loc[["CO2_ANT", "CO2_BIO"], :]
        .sum(axis="index")
        .sort_index(),
        inv_with_pnt_sources.total_emissions.loc["CO2", :].sort_index().fillna(0),
        check_names=False,
    )


def test_speciate_when_profiles_are_there():
    da_speciation = read_speciation_table(
        TESTS_DIR / "speciation" / "table_test_inv_CO2.csv"
    )
    inv = inv_with_pnt_sources.copy()

    inv.set_profile(
        VerticalProfile(ratios=np.array([1.0, 0.0]), height=np.array([1.0, 2.0])),
        category="liku",
        substance="CO2",
    )
    inv_speciated = speciate(inv, substance="CO2", speciation_ratios=da_speciation)
    new_subs = ["CO2_ANT", "CO2_BIO"]
    for sub in new_subs:
        assert sub in inv_speciated.substances

        # Check in the profiles are there
        assert sub in inv_speciated.v_profiles_indexes.coords["substance"].values


def test_speciate_inventory():
    # create a test speciation dictionary
    speciation_dict = {
        ("adf", "CH4"): {
            ("adf", "14CH4"): 0.9,
            ("adf", "12CH4"): 0.1,
        },
        ("liku", "CO2"): {
            ("liku", "14CO2"): 0.5,
            ("liku", "12CO2"): 0.2,
        },
    }
    sp_inv = speciate_inventory(inv_with_pnt_sources, speciation_dict)

    # check that the speciation worked
    assert np.allclose(
        sp_inv.gdf[("adf", "14CH4")], 0.9 * inv_with_pnt_sources.gdf[("adf", "CH4")]
    )
    assert np.allclose(
        sp_inv.gdf[("adf", "12CH4")], 0.1 * inv_with_pnt_sources.gdf[("adf", "CH4")]
    )
    assert np.allclose(
        sp_inv.gdf[("liku", "14CO2")], 0.5 * inv_with_pnt_sources.gdf[("liku", "CO2")]
    )
    assert np.allclose(
        sp_inv.gdf[("liku", "12CO2")], 0.2 * inv_with_pnt_sources.gdf[("liku", "CO2")]
    )
    # check that the original substance is dropped
    assert ("adf", "CH4") not in sp_inv.gdf.columns
    assert ("liku", "CO2") not in sp_inv.gdf.columns
    # check also the point sources
    assert np.allclose(
        sp_inv.gdfs["liku"]["14CO2"],
        0.5 * inv_with_pnt_sources.gdfs["liku"]["CO2"],
    )
    assert np.allclose(
        sp_inv.gdfs["liku"]["12CO2"],
        0.2 * inv_with_pnt_sources.gdfs["liku"]["CO2"],
    )


def test_speciate_inventory_with_profiles():
    # create a test speciation dictionary
    speciation_dict = {
        ("adf", "CH4"): {
            ("adf", "14CH4"): 0.9,
            ("adf", "12CH4"): 0.1,
        },
        ("liku", "CO2"): {
            ("liku", "14CO2"): 0.5,
            ("liku", "12CO2"): 0.2,
        },
    }

    inv = inv_with_pnt_sources.copy()
    inv.set_profile(
        VerticalProfile(ratios=np.array([1.0, 0.0]), height=np.array([1.0, 2.0])),
        category="liku",
        substance="CO2",
    )
    inv.set_profile(
        VerticalProfile(ratios=np.array([0.0, 1.0]), height=np.array([1.0, 2.0])),
        substance="NH3",
    )
    # This is what the indexes should be after setting the profile
    assert inv.v_profiles_indexes.sel(substance="CO2", category="liku").values != -1
    assert inv.v_profiles_indexes.sel(substance="CH4", category="adf").values == -1
    assert inv.v_profiles_indexes.sel(substance="CO2", category="adf").values == -1
    assert inv.v_profiles_indexes.sel(substance="CH4", category="liku").values == -1

    sp_inv = speciate_inventory(inv, speciation_dict)

    assert np.all(
        sp_inv.v_profiles[
            sp_inv.v_profiles_indexes.sel(substance="14CO2", category="liku")
        ].ratios
        == np.array([1.0, 0.0])
    )
    assert np.all(
        sp_inv.v_profiles[
            sp_inv.v_profiles_indexes.sel(substance="12CO2", category="liku")
        ].ratios
        == np.array([1.0, 0.0])
    )

    # Should not be found
    assert sp_inv.v_profiles_indexes.sel(substance="12CO2", category="adf").values == -1

    # NH3 should still be there
    assert sp_inv.v_profiles_indexes.sel(substance="NH3", category="test").values != -1


def test_speciate_nox():
    # create a test inventory containg NOx and 2 test categories
    gdf = gpd.GeoDataFrame(
        {
            ("cat1", "NOx"): np.array([1, 2]),
            ("cat2", "NOx"): np.array([3, 4]),
        },
        geometry=[Point(0, 0), Point(1, 1)],
    )

    inv = Inventory.from_gdf(gdf)
    s_inv = speciate_nox(inv)
    # check that the speciation worked
    assert np.allclose(s_inv.gdf[("cat1", "NO2")], 0.18 * gdf[("cat1", "NOx")])
    # Is a more complex than just a ratio
    assert np.all(s_inv.gdf[("cat1", "NO")] != 0.82 * gdf[("cat1", "NOx")])

    # Test with a dict
    s_inv = speciate_nox(inv, NOX_TO_NO2={"cat1": 0.5, "cat2": 0.2})
    assert np.allclose(s_inv.gdf[("cat1", "NO2")], 0.5 * gdf[("cat1", "NOx")])
    assert np.allclose(s_inv.gdf[("cat2", "NO2")], 0.2 * gdf[("cat2", "NOx")])


def test_simple():

    categories = ["adf", "blek", "liku"]
    da_speciation = xr.DataArray(
        np.array([[1.0, 0.0, 0.5], [0.0, 1.0, 0.5]]),
        coords={
            "substance": ["CO2_ANT", "CO2_BIO"],
            "speciation": range(len(categories)),
            "category": ("speciation", categories),
        },
        dims=["substance", "speciation"],
    )

    inv_speciated = speciate(
        inv_with_pnt_sources, substance="CO2", speciation_ratios=da_speciation
    )


def test_no_speciation_in_ratios():

    da_speciation = xr.DataArray(
        np.array([[1.0, 0.0, 0.5], [0.0, 1.0, 0.5]]),
        coords=[["CO2_ANT", "CO2_BIO"], ["A", "B", "C"]],
        dims=["substance", "category"],
    )
    with pytest.raises(ValueError):
        inv_speciated = speciate(
            inv_with_pnt_sources, substance="CO2", speciation_ratios=da_speciation
        )


def test_merge_substances():

    merged = merge_substances(inv_with_pnt_sources, substances={"GHG": ["CO2", "CH4"]})

    assert "GHG" in merged.substances
    assert "CO2" not in merged.substances
    assert "CH4" not in merged.substances

    # Check that the total emissions are the same
    merged_emissions = merged.total_emissions.loc["GHG"].fillna(0)
    previous_emissions = inv_with_pnt_sources.total_emissions.loc[["CO2", "CH4"]].sum(
        axis="index"
    )
    pd.testing.assert_series_equal(
        merged_emissions.sort_index(),
        previous_emissions.sort_index(),
        check_names=False,
    )


def test_merge_substances_no_drop():

    merged = merge_substances(
        inv_with_pnt_sources, substances={"GHG": ["CO2", "CH4"]}, drop=False
    )

    assert "GHG" in merged.substances
    # Substances should not be dropped
    assert "CO2" in merged.substances
    assert "CH4" in merged.substances


def test_merge_substances_use_as_rename():

    merged = merge_substances(inv_with_pnt_sources, substances={"co2": ["CO2"]})

    # Check that the total emissions are the same
    merged_emissions = merged.total_emissions.loc["co2"]
    previous_emissions = inv_with_pnt_sources.total_emissions.loc["CO2"]
    pd.testing.assert_series_equal(
        merged_emissions.sort_index(),
        previous_emissions.sort_index(),
        check_names=False,
    )


def test_cannot_merge_using_new_substances():

    pytest.raises(
        KeyError,
        merge_substances,
        inv_with_pnt_sources,
        substances={"GHG": ["CO2", "CH4"], "GHG2": ["GHG"]},
    )

    pytest.raises(
        ValueError,
        merge_substances,
        inv_with_pnt_sources,
        substances={"CO2": ["CO2", "CH4"], "GHG2": ["CO2"]},
    )



import numpy as np
import pytest
import xarray as xr

from emiproc.profiles.operators import concatenate_profiles, weighted_combination
from emiproc.profiles.temporal_profiles import (
    AnyProfiles,
    CompositeTemporalProfiles,
    DailyProfile,
    HourOfLeapYearProfile,
    HourOfYearProfile,
    MounthsProfile,
    SpecificDay,
    SpecificDayProfile,
    TemporalProfile,
    WeeklyProfile,
    make_composite_profiles,
)
from emiproc.profiles.utils import merge_indexes
from emiproc.tests_utils.temporal_profiles import (
    daily_test_profile,
    mounths_test_profile,
    weekly_test_profile,
)


@pytest.mark.parametrize(
    "profile_type",
    [
        TemporalProfile,
        DailyProfile,
        WeeklyProfile,
        MounthsProfile,
        HourOfYearProfile,
        HourOfLeapYearProfile,
    ],
)
def test_create_profiles(profile_type):
    profile = profile_type()
    pytest.approx(profile.ratios.sum(axis=1), 1)
    assert profile.ratios.shape == (1, profile.size)


def test_multiple_profiles():
    p = TemporalProfile(
        [
            [0.1, 0.2, 0.3, 0.4],
            [0.1, 0.2, 0.3, 0.4],
            [0.1, 0.2, 0.3, 0.4],
        ],
        size=4,
    )
    assert p.ratios.shape == (3, 4)
    assert p.n_profiles == 3
    assert p.size == 4
    assert len(p) == 3


@pytest.mark.parametrize(
    "profiles_list_list",
    [
        (
            [
                [TemporalProfile(), DailyProfile()],
                [WeeklyProfile()],
            ]
        ),
        (
            [
                [],
                [WeeklyProfile()],
            ]
        ),
        (
            [
                [],
                [],
            ]
        ),
        (
            [
                [],
            ]
        ),
        (
            [
                [TemporalProfile(), DailyProfile()],
                [
                    WeeklyProfile(),
                    SpecificDayProfile(
                        np.array(23 * [0.01] + [0.77]), specific_day=SpecificDay.SUNDAY
                    ),
                ],
            ]
        ),
    ],
)
def test_composite_temporal_profiles(profiles_list_list):
    p = CompositeTemporalProfiles(profiles_list_list)
    assert p.n_profiles == len(profiles_list_list)
    for key, expected in enumerate(profiles_list_list):
        assert len(p[key]) == len(expected)
        for profile, expected_profile in zip(sorted(p[key]), sorted(expected)):
            assert profile == expected_profile

    # Test append
    append_profile = DailyProfile()
    p.append([append_profile])
    assert p.n_profiles == len(profiles_list_list) + 1
    assert len(p[-1]) == 1
    assert p[-1][0] == append_profile
    # Check that the first profile has not changed (assume okay if the others are)
    assert len(p[0]) == len(profiles_list_list[0])
    for profile, expected_profile in zip(sorted(p[0]), sorted(profiles_list_list[0])):
        assert profile == expected_profile

    # Now append specific days
    append_profiles = [
        SpecificDayProfile(
            np.array(23 * [0.02] + [0.54]), specific_day=SpecificDay.SATURDAY
        ),
        SpecificDayProfile(
            np.array(23 * [0.01] + [0.77]), specific_day=SpecificDay.SUNDAY
        ),
    ]
    p.append(append_profiles)
    assert p.n_profiles == len(profiles_list_list) + 2
    assert len(p[-1]) == 2
    for profile, expected_profile in zip(sorted(p[-1]), sorted(append_profiles)):
        assert profile == expected_profile
    # Check that the first profile has not changed (assume okay if the others are)
    assert len(p[0]) == len(profiles_list_list[0])
    for profile, expected_profile in zip(sorted(p[0]), sorted(profiles_list_list[0])):
        assert profile == expected_profile


def test_composite_copy():
    p = CompositeTemporalProfiles(
        [
            [TemporalProfile(), DailyProfile()],
            [],
            [WeeklyProfile()],
        ]
    )
    p2 = p.copy()
    assert p2.n_profiles == p.n_profiles
    for key, expected in enumerate(p):
        assert len(p2[key]) == len(expected)
        for profile, expected_profile in zip(sorted(p2[key]), sorted(expected)):
            assert profile == expected_profile


def test_composite_error_wrong_type():
    with pytest.raises(TypeError):
        CompositeTemporalProfiles(
            [
                [TemporalProfile(), DailyProfile()],
                [WeeklyProfile()],
                [1],
            ]
        )


def test_iterate_over_composite():
    profiles = CompositeTemporalProfiles(
        [
            [TemporalProfile(), DailyProfile()],
            [WeeklyProfile()],
        ]
    )
    for profile in profiles:
        assert isinstance(profile, list)
        for p in profile:
            assert isinstance(p, TemporalProfile)


def test_composite_ratios():
    profiles = CompositeTemporalProfiles(
        [
            [WeeklyProfile(), DailyProfile()],
            [WeeklyProfile()],
        ]
    )
    assert profiles.ratios.shape == (2, 7 + 24)

    # test we can create back from the ratios
    new_profiles = CompositeTemporalProfiles.from_ratios(
        profiles.ratios, profiles.types
    )

    for old, new in zip(profiles, new_profiles):
        assert len(old) == len(new)
        for o, n in zip(sorted(old), sorted(new)):
            assert isinstance(o, TemporalProfile)
            assert isinstance(n, TemporalProfile)
            assert o == n


@pytest.mark.parametrize(
    "ratios, types, expected",
    [
        (
            np.concatenate(
                [
                    (np.ones((2, 7)) / 7.0),
                    (np.ones((2, 24)) / 24.0),
                ],
                axis=1,
            ),
            [WeeklyProfile, DailyProfile],
            {
                0: [WeeklyProfile(), DailyProfile()],
                1: [WeeklyProfile(), DailyProfile()],
            },
        ),
        (
            np.full((2, 7 + 24), np.nan),
            [WeeklyProfile, DailyProfile],
            {0: [], 1: []},
        ),
        (
            np.full((1, 7 + 24), np.nan),
            [WeeklyProfile, DailyProfile],
            {0: []},
        ),
    ],
)
def test_composite_profiles_from_ratios(
    ratios: np.ndarray,
    types: list[TemporalProfile],
    expected: dict[int, list[TemporalProfile]],
):
    profiles = CompositeTemporalProfiles.from_ratios(ratios, types)

    assert profiles.n_profiles == len(expected)
    for index, expected_profiles in expected.items():
        assert len(profiles[index]) == len(expected_profiles)
        for profile, expected_profile in zip(
            sorted(profiles[index]), sorted(expected_profiles)
        ):
            assert profile == expected_profile


def test_internals():
    # This test should not need to exist and could be invalid if someone changes the internal mechanics

    p = CompositeTemporalProfiles(
        [
            [WeeklyProfile(), DailyProfile()],
            [WeeklyProfile()],
        ]
    )
    assert WeeklyProfile in p._profiles
    assert DailyProfile in p._profiles
    assert WeeklyProfile in p._indexes
    assert DailyProfile in p._indexes
    np.testing.assert_array_equal(p._indexes[WeeklyProfile], [0, 1])
    np.testing.assert_array_equal(p._indexes[DailyProfile], [0, -1])
    assert len(p._profiles[WeeklyProfile]) == 2
    assert len(p._profiles[DailyProfile]) == 1


@pytest.mark.parametrize(
    "profiles, expected",
    [
        (
            [
                CompositeTemporalProfiles(
                    [
                        [WeeklyProfile(), DailyProfile()],
                        [WeeklyProfile()],
                    ]
                ),
                CompositeTemporalProfiles(
                    [
                        [WeeklyProfile(), DailyProfile()],
                        [WeeklyProfile()],
                    ]
                ),
                CompositeTemporalProfiles(
                    [
                        [WeeklyProfile()],
                    ]
                ),
            ],
            {
                0: [WeeklyProfile(), DailyProfile()],
                1: [WeeklyProfile()],
                2: [WeeklyProfile(), DailyProfile()],
                3: [WeeklyProfile()],
                4: [WeeklyProfile()],
            },
        ),
        (
            [
                CompositeTemporalProfiles(
                    [
                        [],
                        [],
                    ]
                ),
                CompositeTemporalProfiles(
                    [
                        [],
                        [],
                    ]
                ),
            ],
            {0: [], 1: [], 2: [], 3: []},
        ),
        (
            [
                CompositeTemporalProfiles(
                    [
                        [WeeklyProfile(), DailyProfile()],
                        [WeeklyProfile()],
                    ]
                ),
            ],
            {
                0: [WeeklyProfile(), DailyProfile()],
                1: [WeeklyProfile()],
            },
        ),
    ],
)
def test_join_composites(profiles, expected):
    joined = CompositeTemporalProfiles.join(*profiles)

    for key, expected_profiles in expected.items():
        assert len(joined[key]) == len(expected_profiles)
        for profile, expected_profile in zip(
            sorted(joined[key]), sorted(expected_profiles)
        ):
            assert profile == expected_profile


def test_equality():
    p1 = WeeklyProfile()
    p2 = WeeklyProfile()
    p3 = WeeklyProfile([0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.0])
    p4 = WeeklyProfile([0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.0])
    assert p1 == p2
    assert p1 != p3
    assert p3 == p4


def test_concatenate_composite():
    cat_profiles = concatenate_profiles(
        [
            CompositeTemporalProfiles(
                [
                    [WeeklyProfile(), DailyProfile()],
                    [WeeklyProfile()],
                ]
            ),
            CompositeTemporalProfiles(
                [
                    [WeeklyProfile(), DailyProfile()],
                    [WeeklyProfile()],
                ]
            ),
        ]
    )
    assert cat_profiles.n_profiles == 4
    assert len(cat_profiles[0]) == 2
    assert len(cat_profiles[1]) == 1
    assert len(cat_profiles[2]) == 2
    assert len(cat_profiles[3]) == 1


def test_concatenate_lists():
    cat_profiles = concatenate_profiles(
        [
            [
                [WeeklyProfile(), DailyProfile()],
                [WeeklyProfile()],
            ],
            [
                [WeeklyProfile(), DailyProfile()],
                [WeeklyProfile()],
            ],
        ]
    )
    assert cat_profiles.n_profiles == 4
    assert len(cat_profiles[0]) == 2
    assert len(cat_profiles[1]) == 1
    assert len(cat_profiles[2]) == 2
    assert len(cat_profiles[3]) == 1


def test_wrong_list_concatenate():
    pytest.raises(
        TypeError,
        concatenate_profiles,
        [
            [WeeklyProfile(), DailyProfile()],
            [WeeklyProfile()],
        ],
    )


def test_merging_profiles():
    profiles = AnyProfiles(
        [
            MounthsProfile(
                [
                    [0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.1, 0.2, 0.4, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ],
            ),
            WeeklyProfile([0.1, 0.2, 0.3, 0.1, 0.0, 0.0, 0.3]),
        ]
    )
    assert profiles.n_profiles == 3
    dss = [
        xr.DataArray(
            np.array([[0, 1, 0], [1, -1, 0]]),
            coords={
                "category": ["blek", "liku"],
                "substance": ["CO2", "CH4", "N2O"],
            },
        ).expand_dims(
            {
                "profile": ["MounthsProfile"],
            }
        ),
        xr.DataArray(
            np.array([[2, -1], [2, 2]]),
            coords={
                "category": ["blek", "liku"],
                "substance": ["CO2", "CH4"],
            },
        ).expand_dims(
            {
                "profile": ["WeeklyProfile"],
            }
        ),
    ]

    combined_indexes = merge_indexes(dss)

    make_composite_profiles(profiles, combined_indexes)


def test_weighted_combination():
    weights = np.array([1, 2, 3])
    new_profile = weighted_combination(
        [
            WeeklyProfile([0.1, 0.2, 0.3, 0.4, 0.0, 0.0, 0.0]),
            WeeklyProfile([0.2, 0.1, 0.3, 0.4, 0.0, 0.0, 0.0]),
            WeeklyProfile([0.3, 0.0, 0.3, 0.4, 0.0, 0.0, 0.0]),
        ],
        weights=weights,
    )

    # The combination should give the same as summing the emissions one by one
    assert np.allclose(
        new_profile.ratios.reshape(-1),
        np.array([7 / 30, 2 / 30, 0.3, 0.4, 0.0, 0.0, 0.0]),
    )


if __name__ == "__main__":
    pytest.main([__file__])



"""Information about the categories."""
from emiproc.inventories import EmissionInfo


LARGE_ROAD_TRANSPORT = EmissionInfo(height=0.3, width=7.0)
ZURICH_SOURCES = {
    "c1101_Linienschiffe_Emissionen_Kanton": EmissionInfo(height=1.5, width=5.0),
    "c1102_PrivaterBootsverkehr_Emissionen_Kanton": EmissionInfo(height=0.5),
    "c1201_BahnPersonenverkehr_Emissionen_Kanton": EmissionInfo(height=0.3, width=3.0),
    "c1202_BahnGueterverkehr_Emissionen_Kanton": EmissionInfo(height=0.3, width=3.0),
    "c1203_Tramverkehr_Emissionen_Kanton": EmissionInfo(height=0.3, width=2.0),
    "c1204_Kleinbahnen_Emissionen_Kanton": EmissionInfo(height=0.3, width=2.0),
    "c1301_Personenwagen_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1302_Lastwagen_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1303_Motorraeder_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1304_Linienbusse_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1305_Trolleybusse_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1306_StartStopTankatmung_Emissionen_Kanton": EmissionInfo(height=0.3),
    "c1307_Lieferwagen_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c1308_Reisebusse_Emissionen_Kanton": LARGE_ROAD_TRANSPORT,
    "c2101_Oelheizungen_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2102_Gasheizungen_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2103_HolzheizungenLokalisiert_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2104_HolzheizungenDispers_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2105_Warmwassererzeuger_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2201_BHKW_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c2301_KHKWKehricht_Emissionen_Kanton": EmissionInfo(),
    "c2302_KHKWErdgas_Emissionen_Kanton": EmissionInfo(),
    "c2303_KHKWHeizoel_Emissionen_Kanton": EmissionInfo(),
    "c2401_Klaerschlammverwertung_Emissionen_Kanton": EmissionInfo(
        comment="new junk Cat, group with other category?",
    ),
    "c3101_MaschinenHochbau_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for construction related"
    ),
    "c3102_Bitumen_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for construction related"
    ),
    "c3103_FarbenBaustelle_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for construction related"
    ),
    "c3104_MaschinenTiefbau_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for construction related"
    ),
    "c3105_Strassenbelag_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for construction related"
    ),
    "c3201_Notstromanlagen_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c3301_Prozessenergie_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c3401_Metallreinigung_Emissionen_Kanton": EmissionInfo(),
    "c3402_Holzbearbeitung_Emissionen_Kanton": EmissionInfo(),
    "c3403_Malereien_Emissionen_Kanton": EmissionInfo(),
    "c3404_Textilreinigung_Emissionen_Kanton": EmissionInfo(),
    "c3405_Karosserien_Emissionen_Kanton": EmissionInfo(),
    "c3406_Raeuchereien_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c3407_Roestereien_Emissionen_Kanton": EmissionInfo(height=3.0),
    "c3408_Druckereien_Emissionen_Kanton": EmissionInfo(),
    "c3409_Laboratorien_Emissionen_Kanton": EmissionInfo(),
    "c3410_Bierbrauereien_Emissionen_Kanton": EmissionInfo(),
    "c3411_Brotproduktion_Emissionen_Kanton": EmissionInfo(),
    "c3412_MedizinischePraxen_Emissionen_Kanton": EmissionInfo(),
    "c3413_Gesundheitswesen_Emissionen_Kanton": EmissionInfo(),
    "c3414_Krematorium_Emissionen_Kanton": EmissionInfo(),
    "c3415_Kompostierung_Emissionen_Kanton": EmissionInfo(),
    "c3416_Tankstellen_Emissionen_Kanton": EmissionInfo(),
    "c3417_LoesemittelIG_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for solvents"
    ),
    "c3418_Vergaerwerk_Emissionen_Kanton": EmissionInfo(
        comment="new Cat, group with other category?"
    ),
    "c3419_IndustrielleFZ_Emissionen_Kanton": EmissionInfo(
        comment="added to Agri/Forest vehicel emission",
    ),
    "c4101_ForstwirtschaftlicheFZ_Emissionen_Kanton": EmissionInfo(width=7.0),
    "c4201_LandwirtschaftlicheFZ_Emissionen_Kanton": EmissionInfo(),
    "c4301_Nutzflaechen_Emissionen_Kanton": EmissionInfo(),
    "c4401_Nutztierhaltung_Emissionen_Kanton": EmissionInfo(),
    "c5101_LoesemittelHH_Emissionen_Kanton": EmissionInfo(
        comment="new Cat for solvents"
    ),
    "c5201_Gruenabfallverbrennung_Emissionen_Kanton": EmissionInfo(),
    "c5301_HolzoefenKleingarten_Emissionen_Kanton": EmissionInfo(),
    "c5401_AbfallverbrennungHaus_Emissionen_Kanton": EmissionInfo(),
    "c5501_HausZooZirkustiere_Emissionen_Kanton": EmissionInfo(
        comment="new junk Cat, group with other category?",
    ),
    "c5601_Feuerwerke_Emissionen_Kanton": EmissionInfo(
        comment="new junk Cat, group with other category?",
    ),
    "c5701_Tabakwaren_Emissionen_Kanton": EmissionInfo(),
    "c5801_BrandFeuerschaeden_Emissionen_Kanton": EmissionInfo(
        comment="new junk Cat, group with other category?",
    ),
    "c6101_Waelder_Emissionen_Kanton": EmissionInfo(),
    "c6201_Grasflaechen_Emissionen_Kanton": EmissionInfo(),
    "c6301_Gewaesser_Emissionen_Kanton": EmissionInfo(),
    "c6401_Blitze_Emissionen_Kanton": EmissionInfo(
        comment="new junk Cat, group with other category?",
    ),
}




from datetime import datetime

import pytest

from emiproc import FILES_DIR
from emiproc.exports.hourly import export_hourly_emissions
from emiproc.tests_utils.exports import test_nc_metadata
from emiproc.tests_utils.temporal_profiles import (
    get_random_profiles,
    indexes_inv_catsub,
)
from emiproc.tests_utils.test_inventories import inv

# output path and filename
output_dir = FILES_DIR / "test/exports/test_hourly"
output_dir.mkdir(parents=True, exist_ok=True)


def test_no_profiles():
    """Test that if you have no profile, you get an error."""
    with pytest.raises(ValueError):
        export_hourly_emissions(
            inv=inv,
            path=output_dir,
            netcdf_attributes=test_nc_metadata,
            start_time=datetime(2018, 1, 1),
            end_time=datetime(2018, 1, 2),
        )


def test_export_simple():
    # Export the temporal profiles of TNO
    inv_profiled = inv.copy()
    inv_profiled.set_profiles(
        get_random_profiles(indexes_inv_catsub.max().values + 1),
        indexes=indexes_inv_catsub,
    )
    export_hourly_emissions(
        inv=inv_profiled,
        path=output_dir,
        netcdf_attributes=test_nc_metadata,
        start_time=datetime(2018, 1, 1),
        end_time=datetime(2018, 1, 2),
    )

    # Check that the file is there
    assert (output_dir / "20180101T000000Z.nc").exists()
    # Check until the last
    assert (output_dir / "20180101T230000Z.nc").exists()



"""Test scaling of the emissions."""

from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources


from emiproc.inventories.utils import scale_inventory


scaled_inv = scale_inventory(inv, {"NH3": {"test": 2.42}})


def test_total_emissions_values():

    assert all(scaled_inv.gdf[("test", "NH3")] == 2.42 * inv.gdf[("test", "NH3")])
    assert all(scaled_inv.gdf[("liku", "CO2")] == inv.gdf[("liku", "CO2")])


scaled_inv_pntsrc = scale_inventory(
    inv_with_pnt_sources, {"CO2": {"liku": 1.42}, "NH3": {"test": 2.42}}
)


def test_total_emissions_values_pntsrc():

    assert all(
        scaled_inv_pntsrc.gdf[("test", "NH3")]
        == 2.42 * inv_with_pnt_sources.gdf[("test", "NH3")]
    )
    assert all(
        scaled_inv_pntsrc.gdf[("liku", "CO2")]
        == 1.42 * inv_with_pnt_sources.gdf[("liku", "CO2")]
    )
    assert all(
        scaled_inv_pntsrc.gdf[("adf", "CO2")]
        == inv_with_pnt_sources.gdf[("adf", "CO2")]
    )
    assert all(
        scaled_inv_pntsrc.gdfs["liku"]["CO2"]
        == 1.42 * inv_with_pnt_sources.gdfs["liku"]["CO2"]
    )


def test_scale_with_float():

    scaling_factor = 2.42
    scaled_inv = scale_inventory(inv, scaling_factor)

    assert scaled_inv.total_emissions.equals(inv.total_emissions * scaling_factor)



import numpy as np
import pandas as pd
import xarray as xr

from emiproc.exports.utils import get_temporally_scaled_array
from emiproc.tests_utils import temporal_profiles, test_grids, test_inventories


def test_temporally_scaled_array_on_simple_inv():

    inv = test_inventories.inv.copy()

    profiles = temporal_profiles.three_composite_profiles
    profiles_indexes = temporal_profiles.indexes_inv_catsubcell

    inv.set_profiles(profiles, indexes=profiles_indexes)
    time_range = pd.date_range("2017-12-30", "2018-01-02", freq="h")
    scaled = get_temporally_scaled_array(inv, time_range, sum_over_cells=False)

    assert "time" in scaled.dims
    assert len(scaled.time) == len(time_range)

    assert "category" in scaled.dims
    assert len(scaled.category) == len(inv.categories)

    assert "substance" in scaled.dims
    assert len(scaled.substance) == len(inv.substances)

    # scaled.sum("cell").stack(catsub=["category", "substance"]).plot.line(x="time")



"""Utitlity functions for profiles."""

from __future__ import annotations

import logging
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Type

import numpy as np
import pandas as pd
import xarray as xr

import emiproc
from emiproc.profiles import naming

if TYPE_CHECKING:
    from emiproc.profiles.temporal_profiles import CompositeTemporalProfiles
    from emiproc.profiles.vertical_profiles import VerticalProfiles

logger = logging.getLogger(__name__)


def ratios_to_factors(ratios: np.ndarray) -> np.ndarray:
    """Convert ratios to factors."""

    return ratios * len(ratios)


def factors_to_ratios(factors: np.ndarray) -> np.ndarray:
    """Convert factors to ratios."""

    return factors / len(factors)


def type_in_list(object: Any, objects_list: list[Any]) -> bool:
    """Check if an object of the same type is in the list."""
    return any([isinstance(object, type(o)) for o in objects_list])


def remove_objects_of_type_from_list(object: Any, objects_list: list[Any]) -> list[Any]:
    """Remove objects of the same type from the list."""
    return [o for o in objects_list if not isinstance(object, type(o))]


def get_objects_of_same_type_from_list(
    object: Any, objects_list: list[Any], exact_type: bool = False
) -> list[Any]:
    """Return the object of the same type from the list."""
    func = isinstance if exact_type else lambda x, y: type(x) == y
    return [o for o in objects_list if isinstance(object, type(o))]


def check_valid_indexes(
    indexes: xr.DataArray, profiles: VerticalProfiles | CompositeTemporalProfiles = None
) -> None:
    """Check that the given indexes are valid.

    :raises ValueError: if the indexes are not valid
    """

    if not isinstance(indexes, xr.DataArray):
        raise TypeError(f"Indexes should be an xarray.DataArray, got {type(indexes)=}")
    # Check the dtype of the indexes, should be int
    if indexes.dtype not in [int, np.int64]:
        raise TypeError(f"Indexes should be of type int, got {indexes.dtype=}")

    # check all the dims names are valid
    dims_not_allowed = set(indexes.dims) - set(naming.type_of_dim.keys())
    if len(dims_not_allowed) > 0:
        raise ValueError(
            f"Indexes are not allowed to contain {dims_not_allowed=}, "
            f"allowed dims are {naming.type_of_dim.keys()}"
        )
    # Make sure no coords has duplicated values
    for dim in indexes.dims:
        if indexes.coords[dim].size == 0:
            raise ValueError(f"Indexes are empty for {dim=}")
        if len(indexes.coords[dim]) != len(np.unique(indexes.coords[dim])):
            raise ValueError(
                f"Indexes are not valid, they contain duplicated values for {dim=}:"
                f" {indexes.coords[dim]}"
            )

    if profiles is not None:
        # Check that the max value of the index is given in the profiles
        if indexes.max().values >= len(profiles):
            raise ValueError(
                "Indexes are not valid, they contain values that are not in the"
                f" profiles.Got {indexes.max().values=} but profiles has"
                f" {len(profiles)=}"
            )


def get_desired_profile_index(
    profiles_indexes: xr.DataArray,
    cell: int | None = None,
    cat: str | None = None,
    sub: str | None = None,
    type: str | None = None,
) -> int:
    """Return the index of the desired profile.

    Smart function allowing to select based on desired attributes.
    It will check that the profile can be extracted.
    """

    dimensions_values = {
        "cell": cell,
        "category": cat,
        "substance": sub,
        "type": type,
    }

    # First check that the user did not omit a required dimension
    dims = profiles_indexes.dims
    for dim_name, dim_value in dimensions_values.items():
        if dim_value is None and dim_name in dims:
            raise ValueError(
                f"dimension {dim_name=} is required because the indexes differentiate"
                " it. Please specify it."
            )

    access_dict = {}

    # Add to the access the dimension specified,
    # If a dimension is specified but not in the dims, it means
    # we don't care becausse it is the same for all the dimension cooridnates
    for dim_name, dim_value in dimensions_values.items():
        if dim_value is not None and dim_name in dims:
            if dim_value not in profiles_indexes.coords[dim_name]:
                raise ValueError(
                    f"{dim_name} {dim_value} is not in the profiles indexes, "
                    f"got {profiles_indexes.coords[dim_name]}"
                )
            access_dict[dim_name] = dim_value

    # Access the xarray
    desired_index = profiles_indexes.sel(**access_dict)

    # Check the the seleciton is just a single value
    if desired_index.size != 1:
        raise ValueError(
            f"More than one profile matches the selection: {desired_index}, got"
            f" {desired_index.size =}"
        )

    # Return the index as int
    return int(desired_index.values)


@emiproc.deprecated
def read_profile_csv(
    file: PathLike,
    cat_colname: str = "Category",
    sub_colname: str = "Substance",
    read_csv_kwargs: dict[str, Any] = {},
) -> tuple[pd.DataFrame, str, str | None]:
    """Read a profile csv file and return the dataframe, the category column name and the substance column name.

    Checks the name of the category and substances columns.
    """
    file = Path(file)

    df = pd.read_csv(file, **read_csv_kwargs)
    if cat_colname not in df.columns:
        raise ValueError(f"Cannot find '{cat_colname}' header in {file=}")

    if sub_colname in df.columns:
        sub_header = "Substance"
    else:
        sub_header = None
        logger = logging.getLogger("emiproc.profiles.read_cat_sub_from_csv")
        logger.warning(
            f"Cannot find 'Substance' header in {file=}.\n"
            "All substances will be treated the same way."
        )

    return df, cat_colname, sub_header


def get_profiles_indexes(
    df: pd.DataFrame,
    colnames: dict[str, list[str]] = naming.attributes_accepted_colnames,
) -> xr.DataArray:
    """Return the profiles indexes from the dataframe.

    The dataframe can contain any of the column matching to
    one of the dimensions allowed by the indexes.

    If a column of the index has some nan values, the profiles
    with nan values will fill other other values when no profile is
    given.
    """

    # First get the dimensions present in the columns of the dataframe
    col_of_dim = {}
    for dim, colnames in colnames.items():
        columns = [col for col in colnames if col in df.columns]
        if len(columns) > 1:
            raise ValueError(
                f"Cannot find which column to use for {dim=} in {columns=}.\n"
                f"All columns refer to {dim=}."
            )
        elif len(columns) == 1:
            col_of_dim[dim] = columns[0]
        else:
            logger.debug(f"Cannot find column for {dim=}")
            pass
    logger.info(f"Found {col_of_dim=}")
    # Now get the values of the coords
    coords = {}
    for dim, col in col_of_dim.items():
        coord_values = df[col].unique()
        expected_type = naming.type_of_dim.get(dim, str)
        coord_values = coord_values.astype(expected_type)
        if expected_type == str:
            # Clean the strings from tabs and spaces at the sides
            coord_values = [v.strip() for v in coord_values]
            df[col] = df[col].str.strip()
        coords[dim] = coord_values
    # Create the empty xarray
    indexes = xr.DataArray(
        # -1 means no index specified
        np.full([len(c) for c in coords.values()], -1),
        coords=coords,
        dims=list(coords),
    )
    logger.debug(f"Created {indexes=}")

    # Fill the xarray with the indexes

    indexing_dict = dict(zip(coords, df[list(col_of_dim.values())].values.T))
    indexing_arrays = {}
    for coord, values in indexing_dict.items():
        expected_type = naming.type_of_dim.get(dim, str)
        indexing_arrays[coord] = xr.DataArray(
            values.astype(expected_type), dims=["index"]
        )
    logger.debug(f"Indexing dataarray: {indexing_arrays=}")
    indexes.loc[indexing_arrays] = df.index

    # Finally fill missing profiles with "common profiles" (where there was a nan value)

    for dim in indexes.dims:
        if "nan" in indexes.coords[dim].values:
            common_values = indexes.sel(**{dim: "nan"})
            # Drop the nan indices
            specific_indexes = indexes.sel(
                **{dim: [v for v in indexes.coords[dim].values if v != "nan"]}
            )

            # Get the missing indices
            indexes = xr.where(specific_indexes == -1, common_values, specific_indexes)

    return indexes


def load_country_tz(file: Path | None = None) -> pd.DataFrame:
    """Load the dataframe with the country timezones."""

    if file is None:
        file = Path(*emiproc.__path__).parent / "files" / "country_tz.csv"

    if not file.is_file():
        raise FileNotFoundError(f"Cannot find country_tz {file=}")

    return pd.read_csv(
        file,
        # File start with comment lines starting with '#'
        comment="#",
        # Index column must be 'iso3', the first one
        index_col=0,
        sep=";",
    )


def read_profile_file(file: PathLike, **kwargs: Any) -> pd.DataFrame:
    """Read any kind of profile file and return the dataframe."""
    default_kwargs = {
        "comment": "#",
        "sep": r";|\t|,",
        "engine": "python",  # This is needed to use regex in sep
    }
    for key, value in default_kwargs.items():
        if key not in kwargs:
            kwargs[key] = value
    try:
        logger.log(emiproc.PROCESS, f"Reading {file}")
        df = pd.read_csv(
            file,
            **kwargs,
        )
    except Exception as e:
        raise ValueError(
            f"Could not read profiles from {file}. Please check the format of the file."
        ) from e

    df.rename(columns={col: col.strip() for col in df.columns}, inplace=True)
    # Look a the columns and see if we find columns with names we should clean
    for col in df.columns:
        if col in naming.all_reserved_colnames:
            df[col] = df[col].str.strip()
    return df


def merge_indexes(indexes: list[xr.DataArray]) -> xr.DataArray:
    """Merge together arrays of indexes.

    When merging the indexes, we assume that some arrays have more dimensions
    and thus are more specific than others.
    This implies that the a more specific index will overwrite a less specific one.
    However sometimes conflict can arise. If you have two indexes array and each
    of them are more specific on a dimension.
    Currently function will raise an error in this case.
    """

    # First get the coords of the indexes
    coords = {}
    for index in indexes:
        for coord, values in index.coords.items():
            if coord not in coords:
                coords[coord] = values
            else:
                coords[coord] = np.unique(np.concatenate([coords[coord], values]))
    logger.debug(f"Found {coords=}")

    # Create the empty xarray
    merged_indexes = xr.DataArray(
        # -1 means no index specified
        np.full([len(c) for c in coords.values()], -1),
        coords=coords,
        dims=list(coords),
    )
    logger.debug(f"Created {merged_indexes=}")

    # sort the indexes by the number of dimensions
    indexes.sort(key=lambda x: len(x.dims))

    # Fill the xarray with the indexes
    # Here we assume that the more dimensions we have the more specifc we will be
    # However if the dimensions are others
    specifed_dims = []
    for index_da in indexes:
        # Get the dims that are not specified yet
        dims_to_specify = [dim for dim in index_da.dims if dim not in specifed_dims]
        # Get the dims that have already been specifed but not in this index
        dims_to_overwrite = [dim for dim in specifed_dims if dim not in index_da.dims]
        # Check that there is no conflict
        if len(dims_to_overwrite) > 0:
            raise ValueError(
                f"Cannot merge indexes {indexes=} because of conflict on"
                f" {dims_to_overwrite=}.Please only specify profiles which are being"
                " more specific on the dimensions of others."
            )

        # Get coords where the indexes are given on a single dimension
        indexes_to_set = index_da.stack(z=index_da.dims)
        logger.debug(f"Stacked: {indexes_to_set=}")
        indexes_to_set = indexes_to_set[indexes_to_set != -1]
        indexing_arrays = {}
        for dim in index_da.dims:
            indexing_arrays[dim] = xr.DataArray(indexes_to_set[dim].values, dims="z")
        logger.debug(f"Indexing dataarray: {indexing_arrays=}")
        profile_index = xr.DataArray(indexes_to_set.values, dims="z")
        logger.debug(f"Profiles to set : {profile_index=}")

        # Now fill the indexes xarray
        merged_indexes.loc[indexing_arrays] = profile_index

        logger.debug(f"After Filled: {merged_indexes=}")
        # Add the dims to the specified dims
        specifed_dims.extend(dims_to_specify)

    return merged_indexes


def profiles_to_scalingfactors_dataarray(
    profiles: CompositeTemporalProfiles | VerticalProfiles, indexes: xr.DataArray
) -> xr.DataArray:
    """Convert a profiles object to a ratios DataArray.

    When there is no profile for a given index, the scaling factors are set to 1.0.

    :arg profiles: The profiles object.
    :arg indexes: The indexes DataArray.

    :returns: A DataArray with the scaling factors.
    """

    sf = profiles.scaling_factors
    return xr.DataArray(
        sf[indexes],
        dims=[*indexes.dims, "scaling_factors"],
        coords={
            **indexes.coords,
            "scaling_factors": range(sf.shape[-1]),
        },
        # Remove the profiles with no ratios (will be set to nan)
        # This assumes that no profile = no contribution, so only the other ratios in the cell will have an impact
    ).where(indexes != -1, 1.0)


def ratios_dataarray_to_profiles(
    da: xr.DataArray, rounding_decimals: int | None = None
) -> tuple[np.ndarray, xr.DataArray]:
    """Convert a dataarray of ratios to a profiles array and the indexes compatible for emiproc.

    :arg da: DataArray with the ratios.
        Must contain a 'ratio' dimension. Other dimensions must be the ones
        allowed by the emiproc profiles.
    :arg rounding_decimals: The number of decimals to round the profiles to.
        This can be useful to reduce the number of unique profiles.

    :returns: A tuple with the profiles array and the indexes DataArray.
        The profiles array is a 2D array which can be set at ratios in a Profile object.
        The indexes DataArray is an array that can be set to the indexes of an inventory.

    """
    assert "ratio" in da.dims
    other_coords = {dim: da.coords[dim] for dim in da.dims if dim != "ratio"}

    if len(other_coords) == 0:
        # Add a dummy dimension to allow stacking
        da = da.expand_dims("dummy")
        other_coords = {"dummy": da["dummy"]}

    # Stack the other dimensions
    da_stacked = da.stack(profiles=other_coords.keys())

    # Set the output profiles indexes
    da_profiles_indexes = da_stacked.sum(dim="ratio")
    mask_valid = da_profiles_indexes != 0
    da_profiles_indexes.values = -np.ones(da_profiles_indexes.shape, dtype=int)

    values = da_stacked.sel(profiles=mask_valid).fillna(0.0).values
    if rounding_decimals is not None:
        values = values.round(decimals=rounding_decimals)

    # Get the unique profiles to avoid duplicates
    unique_profiles, unique_indices = np.unique(
        # Fill the nans as the unique functions does not like them
        values,
        axis=-1,
        return_inverse=True,
    )

    # Set the profiles indexes
    da_profiles_indexes.loc[mask_valid] = unique_indices
    profiles_indexes = da_profiles_indexes.unstack(fill_value=-1)

    if "dummy" in profiles_indexes.dims:
        profiles_indexes = profiles_indexes.squeeze("dummy").drop_vars("dummy")

    return unique_profiles.T, profiles_indexes.astype(int)


if __name__ == "__main__":
    print(load_country_tz())



"""Utility functions and constants for emission processing."""

from __future__ import annotations

import json
import logging
import sys
import time
import urllib
from enum import Enum
from functools import cache
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Literal, overload
from urllib.request import urlopen
from warnings import warn
from zipfile import ZipFile

import geopandas as gpd
import numpy as np
import xarray as xr
from shapely.geometry import MultiPolygon, Polygon

from emiproc import FILES_DIR, PROCESS
from emiproc.grids import WGS84, WGS84_PROJECTED, Grid

# constants to convert from yr -> sec
DAY_PER_YR = 365.25
SEC_PER_HOUR = 3600
HOUR_PER_DAY = 24
SEC_PER_DAY = SEC_PER_HOUR * HOUR_PER_DAY
SEC_PER_YR = DAY_PER_YR * SEC_PER_DAY
HOUR_PER_YR = DAY_PER_YR * HOUR_PER_DAY


class Units(Enum):
    """Units for emissions."""

    KG_PER_YEAR = "kg y-1"
    KG_PER_HOUR = "kg h-1"
    KG_PER_M2_PER_S = "kg m-2 s-1"
    MUG_PER_M2_PER_S = "µg m-2 s-1"


PER_M2_UNITS = [Units.KG_PER_M2_PER_S, Units.MUG_PER_M2_PER_S]
PER_CELL_UNITS = [Units.KG_PER_YEAR, Units.KG_PER_HOUR]


def grid_polygon_intersects(
    grid: gpd.GeoSeries,
    poly: Polygon | MultiPolygon,
    within: bool = False,
) -> np.ndarray:
    """Return a mask of the intersection from grid with the polygon.

    This will be fast if the grid is way larger than the polygon.
    Or if the polygon is mostly ouside of the picture.

    :arg grid: The grid to use, as a GeoSeries
    :arg poly: The polygon to check which grid cells intersect with.
    :arg within: If True, return the cells that are fully within the polygon.
        If False, return the cells that intersect with the polygon.
    """
    if isinstance(poly, MultiPolygon):
        return np.logical_or.reduce(
            [grid_polygon_intersects(grid, p, within=within) for p in poly.geoms]
        )

    grid_boundary = Polygon.from_bounds(*grid.total_bounds)

    # poly boundaries
    minx, miny, maxx, maxy = poly.bounds
    sub_grid = grid.cx[minx:maxx, miny:maxy]

    # Crop the part outside of the country
    # TODO: check if this is really useful
    poly_cropped = poly.intersection(grid_boundary)

    out = np.zeros(len(grid), dtype=bool)

    if within:
        mask = sub_grid.within(poly_cropped)
    else:
        mask = sub_grid.intersects(poly_cropped)
    out[sub_grid.index] = mask

    return out


def get_timezones(
    update: bool = False,
    version: str | None = None,
    simplify_tolerance: float = 100,
) -> gpd.GeoDataFrame:
    """Load the timezones shapefile.

    If the file is not found, it will be downloaded from the github repository
    `timezone-boundary-builder <https://github.com/evansiroky/timezone-boundary-builder>`_
    and saved in the files/timezones directory of emiproc.

    :arg update: Download the latest version of the file.
    :arg simplify_tolerance: The tolerance for simplifying the timezones shapes.
        This is used to speed up the calculation and io processes.
        The unit is meters, as we use a projected crs. See
        `geopandas.GeoSeries.simplify <https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoSeries.simplify.html#geopandas.GeoSeries.simplify>`_
        for more information.
        Set this to 0 or False to disable simplification.
    :arg version: The version to download. If not given, the latest version will be downloaded.
        The version correspond to the name of a release (eg: 2023b)

    :returns: The timezones with the shape of each zone.

    """
    logger = logging.getLogger("emiproc.get_timezones")

    latest_dir = FILES_DIR / "timezones" / "latest"
    if version:
        path_to_save = FILES_DIR / "timezones" / version
    else:
        path_to_save = latest_dir
    path_to_save.mkdir(parents=True, exist_ok=True)

    raw_file_path = path_to_save / "combined-shapefile-with-oceans.shp"

    # First get the correct version we need
    if not raw_file_path.is_file() or update:
        repo_url_api = (
            "https://api.github.com/repos/evansiroky/timezone-boundary-builder"
        )
        repo_url = "https://github.com/evansiroky/timezone-boundary-builder"

        # Get the latest version

        if version is None:
            # Get the version to download
            with urllib.request.urlopen(repo_url_api + "/releases/latest") as response:
                if response.status == 200:
                    release_info = json.loads(response.read().decode("utf-8"))
                else:
                    raise ValueError(
                        "Failed to retrieve latest release info. Status code:"
                        f" {response.status}"
                    )

            latest_version = release_info["tag_name"]
            version_to_download = latest_version
        else:
            # Download the specific version
            version_to_download = version

        # Download the file
        url = f"{repo_url}/releases/download/{version_to_download}/timezones-with-oceans.shapefile.zip"
        logger.log(PROCESS, f"Downloading timezones from {url}")
        with urllib.request.urlopen(url) as response:
            if response.status == 200:
                resp = response.read()
            else:
                raise ValueError(
                    "Failed to retrieve latest release info. Status code:"
                    f" {response.status}"
                )
        zipfile = ZipFile(BytesIO(resp))
        zipfile.extractall(path_to_save)

    if not simplify_tolerance:
        return gpd.read_file(str(raw_file_path))

    # Simiplified file
    simplifed_file = raw_file_path.with_stem(
        raw_file_path.stem + f"_simplified_{simplify_tolerance}"
    )
    if not simplifed_file.exists() or update:
        logger.log(PROCESS, f"Simplifying timezones to {simplifed_file}")

        gdf_raw = gpd.read_file(str(raw_file_path))
        # Put on a projected map such that the tolerance is in meters
        gdf_simp = gdf_raw.to_crs(WGS84_PROJECTED).simplify(simplify_tolerance)

        gpd.GeoDataFrame(
            data=gdf_raw[["tzid"]], geometry=gdf_simp.geometry, crs=gdf_simp.crs
        ).to_file(simplifed_file)

    # Load the file
    return gpd.read_file(str(simplifed_file))


def get_timezone_mask(output_grid: Grid, **kwargs) -> np.ndarray:
    """Determine the timezone for each gridcell and return the grid.

    This function calls :py:func:`get_timezones` to get the timezones shapefile.
    Any keyword arguments will be passed to that function.

    Each gridcell gets assigned to the timezone with the most
    area in the cell.

    If for a given grid cell, no timezone is found,
    the timezone UTC is assigned, but the dataset should in theory cover the whole world.

    :arg output_grid: Grid object on which to calculate the timezones.


    :returns:
        Gridded data with the timezone of each gridcell as an array of string
        with the shape of the grid.
    """
    logger = logging.getLogger("emiproc.get_timezone_mask")
    # Put timezones on a projected crs
    gdf = get_timezones(**kwargs)

    grid_gdf = output_grid.gdf.to_crs(gdf.crs)

    # Subselect only the timezones we are interested in
    grid_bounds = grid_gdf.total_bounds
    gdf = gpd.GeoDataFrame(
        gdf.cx[grid_bounds[0] : grid_bounds[2], grid_bounds[1] : grid_bounds[3]]
    )

    # Get the longest timzone names
    longest_name_len = gdf["tzid"].str.len().max()
    timezones = np.full(len(grid_gdf), fill_value="UTC", dtype=f"U{longest_name_len}")

    from emiproc.regrid import calculate_weights_mapping

    weights_mapping = calculate_weights_mapping(grid_gdf, gdf)

    logger.log(PROCESS, "Finished calculated weights mapping")

    mask_completely_in = weights_mapping["weights"] == 1.0

    timezones[weights_mapping["inv_indexes"][mask_completely_in]] = gdf["tzid"].loc[
        weights_mapping["output_indexes"][mask_completely_in]
    ]

    for inv_index in np.unique(weights_mapping["inv_indexes"][~mask_completely_in]):
        # Get the country index wehre it is the largest
        mask_this_index = weights_mapping["inv_indexes"] == inv_index
        this_output_indexes = weights_mapping["output_indexes"][mask_this_index]
        this_weights = weights_mapping["weights"][mask_this_index]
        idx = np.argmax(this_weights)
        timezones[inv_index] = gdf["tzid"].loc[this_output_indexes[idx]]

    return timezones.reshape((output_grid.nx, output_grid.ny))


@cache
def get_natural_earth(
    resolution: str = "10m",
    category: str = "physical",
    name: str = "coastline",
) -> gpd.GeoDataFrame:
    """Download the natural earth file and returns it.

    For more information about the natural earth data, see
    `the natural earth website <https://www.naturalearthdata.com/>`_.

    As this function reads a large dataset, it caches it in case of many uses
    (ex: testing).

    :arg resolution: The resolution for the used shapefile.
        Available resolutions are: '10m', '50m', '110m'
    :arg category: The category of the shapefile.
        Available categories are: 'physical', 'cultural', 'raster'
    :arg name: The name of the shapefile.
        Category of the data to download. Many are availables, look at the
        natural earth website for more info.

    :returns: The shapefile as a GeoDataFrame

    """
    logger = logging.getLogger("emiproc.get_natural_earth")
    path_to_save = FILES_DIR / "natural_earth" / f"ne_{resolution}_{category}_{name}"
    if not path_to_save.exists():
        logger.log(
            PROCESS,
            f"Downloading the natural earth file {resolution}_{category}_{name}",
        )
        URL_TEMPLATE = (
            "https://naturalearth.s3.amazonaws.com/{resolution}_"
            "{category}/ne_{resolution}_{name}.zip"
        )
        path_to_save.mkdir(parents=True, exist_ok=True)
        url = URL_TEMPLATE.format(resolution=resolution, category=category, name=name)
        resp = urlopen(url)
        zipfile = ZipFile(BytesIO(resp.read()))
        zipfile.extractall(path_to_save)

    # Load the country file
    shpfile = str(path_to_save / f"ne_{resolution}_{name}.shp")
    return gpd.read_file(shpfile)


def get_country_mask(
    output_grid: Grid | gpd.GeoSeries,
    resolution: str = "110m",
    weight_filepath: PathLike | None = None,
    return_fractions: bool = False,
) -> np.ndarray | xr.DataArray:
    """Determine the country-code for each gridcell and return the grid.

    Each gridcell gets assigned to code of the country with the most
    area in the cell.

    If for a given grid cell, no country is found (Ocean for example),
    the country-code '-99' is assigned.


    :arg output_grid:
        Contains all necessary information about the output grid.
        Can be a Grid object or a GeoSeries with a custom geometry.
    :arg resolution:
        The resolution for the used shapefile, used as argument for
        :py:func:`get_natural_earth`
    :arg return_fractions:
        In case you want to know the fraction of each country in each grid cell,
        instead of just the main country, set this to True.
        If True, this will return a `xarray.DataArray` with the fraction of each country.
        If False (default), this will return a numpy array with the main country code.


    :returns: Gridded data with the country identifier of each country (eg. BUR).
        Array of 3 char strings of the shape of the grid.
    """
    logger = logging.getLogger("emiproc.get_country_mask")

    if weight_filepath is not None:
        suffix = ".npy" if not return_fractions else ".nc"
        weight_filepath = Path(weight_filepath)
        if weight_filepath.suffix != suffix:
            raise ValueError(
                f"Weight file {weight_filepath} should have suffix {suffix}"
                f" with arg: {return_fractions=}"
            )
        if weight_filepath.is_file():
            try:
                if suffix == ".npy":
                    weigts = np.load(weight_filepath)
                    if weigts.shape == (output_grid.nx, output_grid.ny):
                        return weigts
                    else:
                        logger.warning(
                            f"Weight file {weight_filepath} does not match the grid"
                            " shape, ignoring it."
                        )
                        weight_filepath = None
                elif suffix == ".nc":
                    weigts = xr.load_dataarray(weight_filepath)
                    return weigts
            except Exception as e:
                warn(f"Could not load weight file {weight_filepath}, {e}")
                weight_filepath = None

    if resolution in ["10m", "50m"]:
        logger.log(
            PROCESS,
            f"Computing the country mask with {resolution} resolution.\n"
            "Consider using a coarser resolution to speed up the process "
            "if necessary.",
        )
    start = time.time()

    ne_name = "admin_0_countries"
    countries_gdf = get_natural_earth(
        resolution=resolution, category="cultural", name=ne_name
    )
    if isinstance(output_grid, Grid):
        grid_gdf = output_grid.gdf.copy(deep=True)
    elif isinstance(output_grid, gpd.GeoSeries):
        grid_gdf = gpd.GeoDataFrame(geometry=output_grid)
    else:
        raise TypeError(
            f"output_grid should be a Grid or a GeoSeries, not {type(output_grid)}"
        )

    if grid_gdf.crs != WGS84:
        # make sure the grid is in WGS84 as is the country data
        grid_gdf = grid_gdf.to_crs(WGS84)

    grid_boundary = grid_gdf.geometry.total_bounds

    country_shapes: dict[str, Polygon] = {}
    country_corresponding_codes: list[int] = []
    # 3 char str from ISO, missing value is also -99 so we keep for consistency
    country_mask = np.full(len(grid_gdf), fill_value="-99", dtype="U3")

    # Reduce to the bounds of the grid
    countries_gdf = countries_gdf.cx[
        grid_boundary[0] : grid_boundary[2], grid_boundary[1] : grid_boundary[3]
    ]

    progress = ProgressIndicator(len(countries_gdf) + 10)

    for geometry, iso3, sovereignt in zip(
        countries_gdf.geometry, countries_gdf["ISO_A3_EH"], countries_gdf["SOVEREIGNT"]
    ):
        progress.step()

        mask_intersect = grid_polygon_intersects(grid_gdf.geometry, geometry)
        if np.any(mask_intersect):
            if iso3 == "-99":
                # Countries with missing iso3 code
                logger.info(
                    f"Country {sovereignt=} has no iso3 code (ISO_A3_EH) in natural"
                    f" earth data '{ne_name}'"
                )
                continue
            grid_gdf.loc[grid_gdf.index, iso3] = mask_intersect
            country_shapes[iso3] = geometry
            country_corresponding_codes.append(iso3)
    country_corresponding_codes = np.array(
        country_corresponding_codes,
        # 3 char str
        dtype="U3",
    )

    if return_fractions:
        # Create the output empty array
        da = xr.DataArray(
            np.zeros((len(country_corresponding_codes), len(grid_gdf))),
            coords={
                "country": country_corresponding_codes,
                "cell": grid_gdf.index,
            },
            dims=["country", "cell"],
        )

    # Find how many countries each cell intersected
    progress.step()
    number_of_intersections = grid_gdf[country_shapes.keys()].sum(axis=1)

    # Cells having only one country
    progress.step()
    if not return_fractions:
        # Some small optimization for countries that are in only one cell
        one_cell_df = grid_gdf.loc[number_of_intersections == 1, country_shapes.keys()]
        # Will have in col-0 the index of countries from on_cell_df and col-1 the codes
        cell_country_maps = np.argwhere(one_cell_df.to_numpy())
        country_mask[one_cell_df.index[cell_country_maps[:, 0]]] = (
            country_corresponding_codes[cell_country_maps[:, 1]]
        )
        # Find indexes of cell in more than one country
        progress.step()
        mask_many_cells = number_of_intersections > 1
    else:
        # Take only cells with at least one country
        mask_many_cells = number_of_intersections >= 1
    many_cells_df = grid_gdf.loc[mask_many_cells, country_shapes.keys()]
    if len(many_cells_df) > 0:
        # First column is the index of the cell, second column is the index of the country
        cell_country_duplicates = np.argwhere(many_cells_df.to_numpy())

        # Create two arrays for preparing intersection area between grid cells and countries
        grid_shapes = gpd.GeoSeries(
            grid_gdf.loc[mask_many_cells].geometry.iloc[cell_country_duplicates[:, 0]],
            crs=grid_gdf.crs,
        )
        countries = gpd.GeoSeries(
            np.array([s for s in country_shapes.values()], dtype=object)[
                cell_country_duplicates[:, 1]
            ],
            crs=grid_gdf.crs,
        )
        # Calculate the intersection area
        intersection_shapes = grid_shapes.intersection(countries, align=False)
        # Use projected crs to get correct area
        intersection_areas = intersection_shapes.to_crs(WGS84_PROJECTED).area

        # Prepare a matrix for comparing the area of intersection of cells with each country
        u, i = np.unique(intersection_areas.index, return_inverse=True)

        # rows match each cell that contain duplicate, columns match countries
        areas_matrix = np.zeros(
            (np.max(i) + 1, np.max(cell_country_duplicates[:, 1]) + 1)
        )
        areas_matrix[cell_country_duplicates[:, 0], cell_country_duplicates[:, 1]] = (
            intersection_areas
        )
        if return_fractions:
            # Get the fractions of each country in each cell
            cell_areas = grid_gdf.to_crs(WGS84_PROJECTED).area
            fractions = areas_matrix / cell_areas[u].values[:, None]
            da.loc[
                {
                    "country": country_corresponding_codes,
                    "cell": many_cells_df.index,
                }
            ] = fractions.T
        else:
            # Find the countries in which the area is the largest
            country_mask[many_cells_df.index] = country_corresponding_codes[
                np.argmax(areas_matrix, axis=1)
            ]

    end = time.time()
    logger.log(PROCESS, f"Computation is over, it took {int(end - start)} seconds")
    if return_fractions:
        if weight_filepath is not None:
            da.to_netcdf(weight_filepath)
        return da
    if isinstance(output_grid, Grid):
        country_mask = country_mask.reshape((output_grid.nx, output_grid.ny))
    if weight_filepath is not None:
        np.save(weight_filepath, country_mask)
    return country_mask


def confirm_prompt(question: str) -> bool:
    """Small function to as a prompt with y/n to the user.

    :return: True if the user said y, False otherwise
    """
    reply = None
    while reply not in ("y", "n"):
        reply = input(f"{question} (y/n): ").casefold()
    return reply == "y"


def total_emissions_almost_equal(
    total_dict_1: dict[str, dict[str, float]],
    total_dict_2: dict[str, dict[str, float]],
    relative_tolerance: float = 1e-5,
) -> bool:
    """Test that the total emissions of two inventories are almost equal.

    :arg total_dict_1: The first total emissions dictionary
    :arg total_dict_2: The second total emissions dictionary
    :arg relative_tolerance: The relative tolerance to use for the comparison.
        The comparison is done as follows:
        abs(total_dict_1 - total_dict_2) < relative_tolerance * (total_dict_1 + total_dict_2) / 2

    :returns: True if the total emissions are almost equal, False otherwise
    :raises ValueError: If the two dictionaries have different keys.
    """
    for sub in total_dict_1.keys() | total_dict_2.keys():
        if sub not in total_dict_1 or sub not in total_dict_2:
            raise ValueError(f"Subcategory {sub} is missing in one of the dictionaries")
        for cat in total_dict_1[sub].keys() | total_dict_2[sub].keys():
            if cat not in total_dict_1[sub] or cat not in total_dict_2[sub]:
                raise ValueError(
                    f"Category {cat} is missing in one of the dictionaries for substance {sub}"
                )
            # Get a very small proportion of the total emissions
            err_allowed = (
                relative_tolerance
                * (total_dict_1[sub][cat] + total_dict_2[sub][cat])
                / 2
            )

            if abs(total_dict_1[sub][cat] - total_dict_2[sub][cat]) > err_allowed:
                return False
    return True


class ProgressIndicator:
    """Used to show progress for long operations.

    To not break the progress indicator, make sure there is no
    output to stdout between calls to step()
    """

    steps: int
    curr_step: int

    def __init__(self, steps: int):
        """steps : int
        How many steps the operation takes
        """
        self.curr_step = 0
        self.steps = steps
        # Take the stream from a logger if possible
        logger = logging.getLogger("emiproc.ProgressIndicator")
        logger.setLevel(PROCESS)
        handler = logging.StreamHandler(sys.stdout)
        logger.addHandler(handler)
        self.stream = handler.stream

        self._curr_progress_step = 0

    def step(self):
        """Advance one step, update the progress indicator"""
        self.curr_step += 1
        if (self.curr_step - self._curr_progress_step) / self.steps > 0.001:
            # Show progress when you can update a x.x % value only
            self._show_progress()

    def _show_progress(self):
        self._curr_progress_step = self.curr_step
        progress = self.curr_step / self.steps * 100
        self.stream.write(f"\r{progress:.1f}%")
        if self.curr_step == self.steps:
            self.stream.write("\r")


if __name__ == "__main__":
    progg = ProgressIndicator(100)
    for i in range(100):
        progg.step()
        time.sleep(0.1)



"""Few plot functions for the emiproc package."""

from __future__ import annotations

import itertools
import logging
from os import PathLike
from pathlib import Path
from typing import Any

import geopandas as gpd
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import LogNorm, SymLogNorm

from emiproc.inventories import Inventory
from emiproc.plots import nclcmaps
from emiproc.regrid import get_weights_mapping, weights_remap
from emiproc.utilities import get_natural_earth


def explore_multilevel(gdf: gpd.GeoDataFrame, colum: Any, logscale: bool = False):
    """Explore a multilevel GeodataFrame.

    There is a bug with multilevel datframes that makes them impossible
    to call with the gpd.explore method.
    You can use this instead.
    """
    col_name = str(colum)
    # Deep copy to avoid changing anything
    data = gdf[colum].copy(deep=True)
    if logscale:
        data.loc[data == 0] = np.nan
        data = np.log(data)
    gdf_plot = gpd.GeoDataFrame({col_name: data}, geometry=gdf.geometry)
    return gdf_plot.explore(gdf[colum])


def explore_inventory(
    inv: Inventory, category: None | str = None, substance: None | str = None
):
    """Explore the emission of an inventory."""
    # First check if the data is available
    if inv.gdf is None:
        if category is not None and substance is not None:
            on_others_gdfs = category in inv.gdfs and substance in inv.gdfs[category]
            return inv.gdfs[category].explore()
        # TODO: implement some checks
        raise NotImplementedError()
    elif (
        category is not None
        and category not in inv.gdfs
        and category not in inv.gdf.columns
    ):
        raise IndexError(f"Category '{category}' not in inventory '{inv}'")
    elif (
        substance is not None
        and all((substance not in gdf for gdf in inv.gdfs))
        and substance not in inv.gdf.columns.swaplevel(0, 1)
    ):
        raise IndexError(f"Substance '{substance}' not in inventory '{inv}'")
    elif (
        substance is not None
        and category is not None
        and (category, substance) not in inv.gdf
        and (category not in inv.gdfs or substance not in inv.gdfs[category])
    ):
        raise IndexError(
            f"Substance '{substance}' for Category '{category}' not in inventory '{inv}'"
        )
    else:
        # Valid choice
        pass

    if category is None and substance is None:
        gdf = gpd.GeoDataFrame(
            geometry=pd.concat(
                [inv.geometry, *(gdf.geometry for gdf in inv.gdfs.values())]
            ),
            crs=inv.crs,
        )
        return gdf.explore()
    elif category is not None and substance is None:

        gdf = gpd.GeoDataFrame(
            geometry=pd.concat(
                ([inv.geometry] if category in inv.gdf.columns else [])
                + ([inv.gdfs[category].geometry] if category in inv.gdfs else [])
            ),
            crs=inv.crs,
        )
        return gdf.explore()
    elif category is not None and substance is not None:
        on_main_grid = (category, substance) in inv.gdf.columns
        on_others_gdfs = category in inv.gdfs and substance in inv.gdfs[category]
        gdf = gpd.GeoDataFrame(
            {
                str((category, substance)): pd.concat(
                    ([inv.gdf[(category, substance)]] if on_main_grid else [])
                    + ([inv.gdfs[category][substance]] if on_others_gdfs else [])
                )
            },
            geometry=pd.concat(
                ([inv.geometry] if on_main_grid else [])
                + ([inv.gdfs[category].geometry] if on_others_gdfs else [])
            ),
            crs=inv.crs,
        )
        return gdf.explore(gdf[str((category, substance))])
    else:
        raise NotImplementedError()


def plot_inventory(
    inv: Inventory,
    figsize=(16, 9),
    q=0.001,
    vmin: None | float = None,
    vmax: None | float = None,
    cmap=nclcmaps.cmap("WhViBlGrYeOrRe"),
    symcmap="RdBu_r",
    spec_lims: None | tuple[float] = None,
    out_dir: PathLike | None = None,
    axis_formatter: str | None = None,
    x_label="lon [°]",
    y_label="lat [°]",
    add_country_borders: bool = False,
    total_only: bool = False,
    reverse_y: bool = False,
):
    """Plot an inventory.

    Will plot all the combination of substnaces/category of the inventory.
    Will also plot the total emission for each substance.

    :arg axis_formatter: for example "{x:6.0f}" will show 6 number and 0
        after the . , which is useful for swiss coordinates.
    """

    logger = logging.getLogger(__name__)

    grid = inv.grid
    grid_shape = (grid.nx, grid.ny)

    def get_vmax(data: np.ndarray) -> float:
        if vmax is not None:
            return vmax
        return np.quantile(data, 1 - q)

    def get_vmin(data: np.ndarray) -> float:
        if vmin is not None:
            return vmin
        return np.quantile(data, q)

    def get_norm_and_cmap(data: np.ndarray) -> tuple[mpl.colors.Normalize, str]:
        if np.any(data < 0):
            abs_values = np.abs(data)
            vmax_ = get_vmax(abs_values)
            vmin_ = get_vmin(abs_values)
            # Use symlog instead
            return SymLogNorm(linthresh=vmin_, vmin=-vmax_, vmax=vmax_), symcmap
        else:
            return LogNorm(vmin=get_vmin(data), vmax=get_vmax(data)), cmap

    if len(inv.categories) == 1:
        logger.info("Only one category, will plot only the total emissions")
        total_only = True

    lon_range = grid.lon_range if hasattr(grid, "lon_range") else np.arange(grid.nx)
    lat_range = grid.lat_range if hasattr(grid, "lat_range") else np.arange(grid.ny)

    x_min, x_max = lon_range[0], lon_range[-1]
    y_min, y_max = lat_range[0], lat_range[-1]

    if add_country_borders and (
        not hasattr(grid, "lat_range") or hasattr(grid, "lon_range")
    ):
        raise ValueError(
            "Cannot add country borders without grid lat_range and lon_range"
        )
    elif add_country_borders:
        gdf_countries = get_natural_earth(
            resolution="10m", category="cultural", name="admin_0_countries"
        )
        # Crop the countries to the grid
        gdf_countries = gdf_countries.cx[x_min:x_max, y_min:y_max].clip_by_rect(
            x_min, y_min, x_max, y_max
        )

        def add_country_borders(ax: mpl.axes.Axes):
            gdf_countries.boundary.plot(ax=ax, color="black", linewidth=0.5)

    else:
        add_country_borders = lambda ax: None

    def add_ax_info(ax: mpl.axes.Axes):
        if axis_formatter is not None:
            ax.xaxis.set_major_formatter(axis_formatter)
            ax.yaxis.set_major_formatter(axis_formatter)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        if spec_lims:
            ax.set_xlim(spec_lims[0], spec_lims[1])
            ax.set_ylim(spec_lims[2], spec_lims[3])

    if out_dir is not None:
        plt.ioff()
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

    per_substances_per_sector_emissions = {}
    for sub in inv.substances:
        per_sector_emissions = {}
        per_substances_per_sector_emissions[sub] = per_sector_emissions
        total_sub_emissions = np.zeros(grid_shape).T
        for cat in inv.categories:
            if (cat, sub) not in inv.gdf:
                # TODO: this will miss point sources for the total_sub_emissions
                # And also miss point sources for that category
                print(f"passsed {sub},{cat} no data")
                continue
            emissions = inv.gdf[(cat, sub)].copy(deep=True).to_numpy()
            if cat in inv.gdfs and sub in inv.gdfs[cat]:
                if out_dir:
                    weights_file = out_dir / f".emiproc_weights_{inv.name}_gdfs_{cat}"
                else:
                    weights_file = None
                weights_mapping = get_weights_mapping(
                    weights_file, inv.gdfs[cat], inv.gdf, loop_over_inv_objects=True
                )

                gdfs_emissions = weights_remap(
                    weights_mapping, inv.gdfs[cat][sub], len(emissions)
                )
                # TODO: could check and plot only the point sources as well
                emissions += gdfs_emissions

            per_sector_emissions[cat] = np.sum(emissions)

            # from ha to m2
            emissions /= inv.cell_areas

            y_slice = slice(None, None, 1 if reverse_y else -1)
            emissions = emissions.reshape(grid_shape).T[y_slice, :]

            total_sub_emissions += emissions

            if not np.any(emissions):
                print(f"passsed {sub},{cat} no emissions")
                continue

            if total_only:
                continue

            fig, ax = plt.subplots(
                figsize=figsize,
                # gridspec_kw={"right": 0.9, "wspace": 0},
            )

            # Check if there are values below 0
            emission_non_zero_values = emissions[
                (emissions != 0) & (~np.isnan(emissions))
            ]
            if len(emission_non_zero_values) == 0:
                print(f"passsed {sub},{cat} no emissions")
                continue

            norm, this_cmap = get_norm_and_cmap(emission_non_zero_values)

            im = ax.imshow(
                emissions,
                norm=norm,
                cmap=this_cmap,
                extent=[x_min, x_max, y_min, y_max],
            )
            add_ax_info(ax)
            ax.set_title(f"{sub} - {cat}: " f"{per_sector_emissions[cat]:.2} " f"kg/y")
            fig.colorbar(
                im,
                label="kg/y/m2",
                extend="both",
                extendfrac=0.02,
                # cax=cax,
            )

            add_country_borders(ax)

            fig.tight_layout()

            if out_dir:
                file_name = Path(out_dir) / f"raster_{sub}_{cat}"

                fig.savefig(file_name.with_suffix(".png"))
                # fig.savefig(file_name.with_suffix(".pdf"))
                fig.clear()
            else:
                plt.show()

            plt.close(fig)

        if not np.any(total_sub_emissions):
            print(f"passsed {sub},total_emissions, no emissions")
            continue

        fig, ax = plt.subplots(figsize=figsize)

        emission_non_zero_values = total_sub_emissions[
            (total_sub_emissions > 0) & (~np.isnan(total_sub_emissions))
        ]
        if len(emission_non_zero_values) == 0:
            print(f"passsed {sub},total_emissions, no emissions")
            continue

        norm, this_cmap = get_norm_and_cmap(emission_non_zero_values)

        im = ax.imshow(
            total_sub_emissions,
            norm=norm,
            cmap=this_cmap,
            extent=[x_min, x_max, y_min, y_max],
        )
        ax.set_title(
            f"Total {sub}: " f"{sum(per_sector_emissions.values()):.2} " f"kg/y"
        )
        fig.colorbar(
            im,
            label="kg/y/m2",
            extend="both",
            extendfrac=0.02,
            # cax=cax,
        )

        add_ax_info(ax)
        add_country_borders(ax)
        fig.tight_layout()

        if out_dir:
            file_name = Path(out_dir) / f"raster_total_{sub}"

            fig.savefig(file_name.with_suffix(".png"))
            # fig.savefig(file_name.with_suffix(".pdf"))
            fig.clear()
        else:
            plt.show()

        plt.close(fig)

    # A bar plot of the total emissions for each substances and each category
    sorted_categories = sorted(inv.categories)
    n_substances = len(inv.substances)
    fig, axes = plt.subplots(
        figsize=(len(inv.categories) * 0.5, n_substances),
        nrows=n_substances,
        sharex=True,
    )

    color_iter = itertools.cycle(plt.rcParams["axes.prop_cycle"])
    colors_of_categories = {cat: next(color_iter)["color"] for cat in sorted_categories}
    for i, sub in enumerate(inv.substances):
        if n_substances > 1:
            ax = axes[i]
        else:
            ax = axes
        for j, cat in enumerate(sorted_categories):
            ax.bar(
                j,
                per_substances_per_sector_emissions[sub].get(cat, 0),
                color=colors_of_categories.get(cat, "black"),
            )
        ax.set_ylabel(f"{sub} [kg/y]")

    # Add ticks on the last ax (the one at the bottom)
    ax.set_xticks(range(len(sorted_categories)))
    ax.set_xticklabels(sorted_categories, rotation=45, ha="right")

    if out_dir:
        file_name = Path(out_dir) / f"barplot_total_emissions"
        fig.savefig(file_name.with_suffix(".png"))
        fig.clear()
    else:
        plt.show()

    plt.close(fig)



from emiproc.tests_utils import temporal_profiles, test_inventories
from emiproc.inventories.utils import group_categories


def test_group_profiles_with_time_profiles():

    inv = test_inventories.inv.copy()

    inv.set_profiles(
        temporal_profiles.three_composite_profiles,
        indexes=temporal_profiles.indexes_inv_catsubcell,
    )

    groupped = group_categories(inv, {"all": inv.categories})

    assert groupped.categories == ["all"]

    assert groupped.t_profiles_indexes["category"].values == ["all"]



from emiproc.exports.netcdf import nc_cf_attributes

test_nc_metadata = nc_cf_attributes(
    author="Emiproc Test",
    contact="name@domain.com",
    title="Hourly Emissions for ...",
    source="emiproc tests",
    institution="emiproc, tests",
)



"""Zurich inventory."""
from os import PathLike
from pathlib import Path
from emiproc.inventories import Inventory, Substance, Category
from emiproc.inventories.utils import list_categories, process_emission_category
from shapely.geometry import Point


class MapLuftZurich(Inventory):
    """Inventory of Zurich based on the mapluft.gbd file.

    It contains only shaped emissions for each categories.
    """

    mapluft_gdb: Path

    def __init__(
        self,
        mapluft_gdb: PathLike,
        # Hardcoded the name of the substances for simplicity
        substances: list[Substance] = [
            "CO2",
            "CO",
            "PM10ex",
            "PM10non",
            "PM25ex",
            "PM25non",
            "SO2",
            "NOx",
            "N2O",
            "NH3",
            "CH4",
            "BC",
            "VOC",
            "benzene",
        ],
        categories: list[Category] = [],
        remove_josefstrasse_khkw: bool = False,
        convert_lines_to_polygons: bool = True,
    ) -> None:
        """Load the mapluft inventory.

        :arg mapluft_gdb: The Mapluft file
        :arg substances: A list of substances to load.
            (by default all of them).
            Categories not contianing any of the substances are not loaded.
        :arg categories: A list of categories to load (if one is interested
            in only a subset).
            If not specified, all categories are loaded.
        :arg remove_josefstrasse_khkw: Whether the incineration plant
            at josefstrasse should be removed from the inventory.
            Emission for category 'c2301_KHKWKehricht_Emissionen_Kanton' at 
            the Josefstrasse location will be removed from the inventory.
            It should be planned to be removed in March 2021.
            The other Josefstrasse category will still be present, as they
            account for some kinds of energy production.
            In case  remove_josefstrasse_khkw, the emissions are not set to any
            other location in the inventory.
        :arg convert_lines_to_polygons: Whether this should convert line emissions 
            to polygons. Only few models can handle line emissions.
            The default width of the line is 10m for all categories.
            This is not currently not changeable.
        """
        super().__init__()
        self.mapluft_gdb = Path(mapluft_gdb)

        if not self.mapluft_gdb.is_dir():
            raise FileNotFoundError(f"{self.mapluft_gdb=} is not a existing directory.")

        if not categories:
            # Load all the categories
            categories = list_categories(mapluft_gdb)
        else:
            self.history.append("Only a subset of categories was loaded.")

        # Maps substances name from the mapluft files to the names in emiproc
        emission_names = {
            (f"Emission_{sub}" if sub != "benzene" else "Emission_Benzol"): sub
            for sub in substances
        }

        # Mapluft has no grid
        self.gdf = None

        self.gdfs = {}

        for category in categories:
            gdf = process_emission_category(
                mapluft_gdb, category, convert_lines_to_polygons=convert_lines_to_polygons
            )
            # Select the columns with the emissions values
            names_in_gdf = [
                name for name in emission_names.keys() if name in gdf.columns
            ]
            if not names_in_gdf:
                # Ignore categories not containing any substance of interest
                continue
            gdf = gdf.loc[:, list(names_in_gdf) + ["geometry"]]

            
            if remove_josefstrasse_khkw and category == "c2301_KHKWKehricht_Emissionen_Kanton":
                # Check point sources at the josefstrasse location
                mask_josefstrasse = gdf.geometry == Point(2681839.000, 1248988.000)
                if any(mask_josefstrasse):
                    self.history.append(
                        f"Removed {sum(mask_josefstrasse)} point source"
                        f"from josefstrasse of {category=}."
                    )
                gdf = gdf.loc[~mask_josefstrasse]
            # Only keep the substances
            self.gdfs[category] = gdf.rename(columns=emission_names, errors="ignore")


if __name__ == "__main__":

    from emiproc.inventories.zurich.categories_info import ZURICH_SOURCES
    mapluft_file = Path("/store/empa/em05/mapluft/mapLuft_2020_v2021.gdb")
    inv = MapLuftZurich(mapluft_file)
    inv.emission_infos = ZURICH_SOURCES
    print(inv.emission_infos)



"""Comparison of Swiss emissions in 2020 and 2022."""

# %% Imports
from pathlib import Path

from emiproc.inventories.swiss import SwissRasters
from emiproc.inventories.utils import drop
from emiproc.speciation import merge_substances

# %% Select the path with my data
data_path = Path(r"C:\Users\coli\Documents\ZH-CH-emission\Data\CHEmissionen")


# %% Create the inventory object
ch_22 = SwissRasters(
    data_path=data_path / "CH_Emissions_2015_2020_2022_CO2_CO2biog_CH4_N2O_BC_AP.xlsx",
    rasters_dir=data_path / "ekat_gridascii",
    rasters_str_dir=data_path / "ekat_str_gridascii_v_swiss2icon",
    requires_grid=False,
    year=2022,
)
ch_20 = SwissRasters(
    data_path=data_path / "CH_Emissions_2015_2020_2022_CO2_CO2biog_CH4_N2O_BC_AP.xlsx",
    rasters_dir=data_path / "ekat_gridascii",
    rasters_str_dir=data_path / "ekat_str_gridascii_v_swiss2icon",
    requires_grid=False,
    year=2020,
)
ch_20 = drop(ch_20, substances=["CO2", "CO2_biog"], keep_instead_of_drop=True)
ch_20 = merge_substances(ch_20, {"CO2": ["CO2", "CO2_biog"]}, inplace=True)
ch_22 = drop(ch_22, substances=["CO2", "CO2_biog"], keep_instead_of_drop=True)
ch_22 = merge_substances(ch_22, {"CO2": ["CO2", "CO2_biog"]}, inplace=True)



"""Convert the TNO inventory to a netcdf file on a regular raster grid.

By scaling the emissions with the time profiles, we export hourly emissions values.
"""
# %%
%load_ext autoreload
%autoreload 2
# %%
from datetime import datetime
from pathlib import Path

from emiproc.inventories.tno import TNO_Inventory
from emiproc.regrid import remap_inventory

from emiproc.utilities import Units
from emiproc.grids import RegularGrid
from emiproc.inventories.utils import group_categories

from emiproc.exports.netcdf import nc_cf_attributes
from emiproc.exports.hourly import export_hourly_emissions

from emiproc import FILES_DIR


# %%
# path to input inventory
input_path = FILES_DIR / r"test\tno\tno_test_minimal.nc"
profiles_dir = FILES_DIR / r"profiles\tno"
# output path and filename
output_dir = input_path.parent
output_path = output_dir / f"{input_path.stem}_with_hourly_emissions"

# %%
inv = TNO_Inventory(
    input_path,
    # Take only CH4
    substances_mapping={"co2_ff": "CO2","co2_bf": "CO2","ch4": "CH4"},
    profiles_dir=profiles_dir
)



# %% Regular grid domain, If you want to use the TNO grid, use output_grid = inv.grid


output_grid = RegularGrid(
    xmin=-40, ymin=20, xmax=80, ymax=80, nx=312, ny=208
)


remaped_tno = remap_inventory(
    inv=inv,
    grid=output_grid,
)
# %%


nc_metadata = nc_cf_attributes(
    author="LastName FirstName",
    contact="name@domain.com",
    title="Hourly Emissions for ...",
    source=input_path.name,
    institution="Instiution, Country",
)


# %% Export the temporal profiles of TNO
export_hourly_emissions(
    inv=remaped_tno,
    path=output_dir,
    netcdf_attributes=nc_metadata,
    start_time=datetime(2018, 1, 1),
    end_time=datetime(2018, 1, 2),
)



import geopandas as gpd

from shapely.geometry import Point, Polygon

from emiproc.grids import RegularGrid, GeoPandasGrid

# Regular grid for testing
# Note that this grid is big enough to include the inventories from the
# test_utils/test_inventories.py module
regular_grid = RegularGrid(
    xmin=-1, xmax=5, ymin=-2, ymax=3, nx=10, ny=15, name="Test Regular Grid"
)

basic_serie = gpd.GeoSeries(
    [
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
        Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
        Polygon(((2, 1), (2, 2), (3, 2), (3, 1))),
    ]
)

# Another basic serie, on which it is intersting to remap the first serie
basic_serie_2 = gpd.GeoSeries(
    [
        Polygon(((0.75, 0.5), (0.75, 1.5), (1.75, 1.5), (1.75, 0.5))),
        Polygon(((1.75, 0.5), (1.75, 1.5), (2.75, 1.5), (2.75, 0.5))),
    ]
)
basic_grid = GeoPandasGrid(gpd.GeoDataFrame(geometry=basic_serie))
basic_grid_2 = GeoPandasGrid(gpd.GeoDataFrame(geometry=basic_serie_2))

basic_serie_of_size_2 = gpd.GeoSeries(
    [
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
    ]
)

gpd_grid = GeoPandasGrid(
    gpd.GeoDataFrame(
        geometry=[
            Polygon(((0.5, 0.5), (0.5, 1.5), (1.5, 1.5))),
            Polygon(((0.5, 0.5), (1.5, 0.5), (1.5, 1.5))),
            Polygon(((2.5, 0.5), (1.5, 1.5), (1.5, 0.5))),
            Polygon(((2.5, 0.5), (2.5, 1.5), (1.5, 1.5))),
        ]
    )
)

# A regular grid over some african countries
# It makes sure some grid cells are in one country only
# some countries are too small for country mask (Gambia (GMB))
# Some cells are ocean
regular_grid_africa = RegularGrid(
    # 19°24'03.0"N 20°30'34.3"W
    # 4°39'55.8"N 9°14'11.6"W
    xmin=-20.5,
    xmax=-9.25,
    ymin=4.65,
    ymax=20.9,
    nx=10,
    ny=10,
)



"""Test the speciation module."""

from emiproc.inventories.utils import drop
from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources


def test_inv_with_gdfs():

    categories = ["adf", "blek", "liku"]


    inv_dropped = drop(
        inv_with_pnt_sources, substances=["CO2"], categories=categories
    )


    # Check the the categories are dropped
    assert all([cat not in inv_dropped.categories for cat in categories])
    # Was kept
    assert "other" in inv_dropped.categories

    assert "CO2" not in inv_dropped.substances
    assert "NH3" in inv_dropped.substances
    assert "AITS" in inv_dropped.substances


def test_inv_drop_cat():

    categories = ["adf"]

    inv_dropped = drop(
        inv, categories=categories
    )

    assert "adf" not in inv_dropped.categories
    assert "liku" in inv_dropped.categories
    # Check we did not change the original inventory
    assert "adf" in inv.categories


def test_inv_drop_sub():

    subtances = ["CO2"]

    inv_dropped = drop(
        inv, substances=subtances
    )

    assert "CO2" not in inv_dropped.substances
    assert "NH3" in inv_dropped.substances
    # Check we did not change the original inventory
    assert "CO2" in inv.substances


def test_keep_instead_of_drop():
    subtances = ["CO2"]

    inv_dropped = drop(
        inv, substances=subtances, keep_instead_of_drop=True
    )

    assert "CO2" in inv_dropped.substances
    assert "NH3" not in inv_dropped.substances
    # Check we did not change the original inventory
    assert "CO2" in inv.substances



"""Test the addition of two inventories."""

from __future__ import annotations

import pytest

from emiproc.inventories.utils import add_inventories, scale_inventory
from emiproc.profiles.temporal_profiles import HourOfYearProfile, WeeklyProfile
from emiproc.tests_utils import temporal_profiles, test_inventories


def test_self_addition():
    """Test the addition of an inventory with itself."""

    inv = test_inventories.inv_with_pnt_sources
    inv_added = add_inventories(inv, inv)

    assert inv_added.total_emissions.loc[(inv.substances, inv.categories)].equals(
        scale_inventory(inv, 2).total_emissions.loc[(inv.substances, inv.categories)]
    )


def test_addition():
    """Test the addition of two inventories."""

    inv1 = test_inventories.inv_with_pnt_sources
    inv2 = test_inventories.inv

    inv_added = add_inventories(inv1, inv2)

    # Here we need to fill the nan values with 0 to be able to compare the dataframes
    tot_added = inv_added.total_emissions.fillna(0)
    # Also expand the dataframes to have the same columns and rows
    tot1 = inv1.total_emissions.reindex_like(tot_added).fillna(0)
    tot2 = inv2.total_emissions.reindex_like(tot_added).fillna(0)

    assert tot_added.equals(tot1 + tot2)


def test_cannot_add_different_grid():
    """Test that we can only add inventories on the same grid."""

    inv1 = test_inventories.inv_on_grid_serie2_bis
    inv2 = test_inventories.inv_on_grid_serie2

    pytest.raises(ValueError, add_inventories, inv1, inv2)


def test_profiles():
    """Test the addition of two inventories with profiles."""

    inv1 = test_inventories.inv.copy()
    inv2 = test_inventories.inv.copy()

    inv1.set_profiles(
        temporal_profiles.three_composite_profiles,
        indexes=temporal_profiles.indexes_inv_catsubcell,
    )
    inv2.set_profiles(
        temporal_profiles.get_random_profiles(
            temporal_profiles.indexes_inv_catsub_missing.max().values + 1,
            profile_types=[HourOfYearProfile, WeeklyProfile],
        ),
        indexes=temporal_profiles.indexes_inv_catsub_missing,
    )

    summed_inv = add_inventories(inv1, inv2)
